"""Test distance clustering statistics."""

import numpy as np
import pytest

from psifr import stats


@pytest.fixture()
def data():
    """Create position data for a list."""
    list_data = {
        'pool_position': [0, 1, 2, 3, 4, 5, 6, 7],
        'recall_position': [3, 2, 1, 7, 0, 6, 5],
    }
    return list_data


@pytest.fixture()
def list_data():
    """Create list data with item and category information."""
    data = {
        'pool_items': [[0, 1, 2, 3, 4, 5, 6, 7]],
        'recall_items': [[3, 2, 1, 7, 0, 6, 5]],
        'pool_test': [[1, 1, 2, 2, 2, 2, 1, 1]],
        'recall_test': [[2, 2, 1, 1, 1, 1, 2]],
    }
    return data


@pytest.fixture()
def distance():
    """Create a distance matrix."""
    mat = np.array(
        [
            [0, 1, 1, 1, 2, 2, 2, 2],
            [1, 0, 1, 1, 2, 2, 2, 2],
            [1, 1, 0, 1, 2, 2, 2, 2],
            [1, 1, 1, 0, 2, 2, 2, 2],
            [2, 2, 2, 2, 0, 3, 3, 3],
            [2, 2, 2, 2, 3, 0, 3, 3],
            [2, 2, 2, 2, 3, 3, 0, 3],
            [2, 2, 2, 2, 3, 3, 3, 0],
        ]
    )
    return mat


def test_distance_count(data, distance):
    """Test distance bin count."""
    edges = [0.5, 1.5, 2.5, 3.5]
    actual, possible = stats.count_distance(
        distance,
        edges,
        [data['pool_position']],
        [data['recall_position']],
        [data['pool_position']],
        [data['recall_position']],
        count_unique=False,
    )
    # recalls: [3, 2, 1, 7, 0, 6, 5]
    # actual: [1, 1, 2, 2, 2, 3]
    # possible: [
    #     [1, 1, 1, 2, 2, 2, 2],
    #     [1, 1, 2, 2, 2, 2],
    #     [1, 2, 2, 2, 2],
    #     [2, 3, 3, 3],
    #     [2, 2, 2],
    #     [3, 3]
    # ]
    expected_actual = np.array([2, 3, 1])
    expected_possible = np.array([6, 16, 5])
    np.testing.assert_array_equal(actual.to_numpy(), expected_actual)
    np.testing.assert_array_equal(possible.to_numpy(), expected_possible)


def test_distance_count_unique(data, distance):
    """Test distance bin count by transition."""
    edges = [0.5, 2.5, 3.5]
    inputs = [
        distance,
        edges,
        [data['pool_position']],
        [data['recall_position']],
        [data['pool_position']],
        [data['recall_position']],
    ]

    # first check these bins with count_unique=False
    actual, possible = stats.count_distance(*inputs, count_unique=False)
    np.testing.assert_array_equal(actual.to_numpy(), np.array([5, 1]))
    np.testing.assert_array_equal(possible.to_numpy(), np.array([22, 5]))

    # now test with only one bin increment per transition
    actual, possible = stats.count_distance(*inputs, count_unique=True)
    np.testing.assert_array_equal(actual.to_numpy(), np.array([5, 1]))
    np.testing.assert_array_equal(possible.to_numpy(), np.array([5, 2]))


def test_rank_distance(list_data, distance):
    """Test rank distance measure."""
    ranks = stats.rank_distance(
        distance,
        list_data['pool_items'],
        list_data['recall_items'],
        list_data['pool_items'],
        list_data['recall_items'],
    )
    # recalls: [3, 2, 1, 7, 0, 6, 5]
    # actual: [1, 1, 2, 2, 2, 3]
    # possible: [
    #     [1, 1, 1, 2, 2, 2, 2],
    #     [1, 1, 2, 2, 2, 2],
    #     [1, 2, 2, 2, 2],
    #     [2, 3, 3, 3],
    #     [2, 2, 2],
    #     [3, 3],
    # ]
    # actual_rank: [2.0, 1.5, 3.5, 1.0, 2.0, 1.5]
    # possible_rank: [
    #     [2.0, 2.0, 2.0, 5.5, 5.5, 5.5, 5.5],
    #     [1.5, 1.5, 4.5, 4.5, 4.5, 4.5],
    #     [1.0, 3.5, 3.5, 3.5, 3.5],
    #     [1.0, 3.0, 3.0, 3.0],
    #     [2.0, 2.0, 2.0],
    #     [1.5, 1.5],
    # ]
    expected = np.array(
        [
            1 - (1 / 6),
            1 - (0.5 / 5),
            1 - (2.5 / 4),
            1 - (0 / 3),
            1 - (1 / 2),
            1 - (0.5 / 1),
        ]
    )
    np.testing.assert_allclose(ranks, expected)


def test_rank_distance_within(list_data, distance):
    """Test rank distance for within-category transitions."""
    ranks = stats.rank_distance(
        distance,
        list_data['pool_items'],
        list_data['recall_items'],
        list_data['pool_items'],
        list_data['recall_items'],
        pool_test=list_data['pool_test'],
        recall_test=list_data['recall_test'],
        test=lambda x, y: x == y,
    )
    expected = np.array([1 - (0 / 2), 1 - (1.5 / 2), 1 - (0 / 1), np.nan])
    np.testing.assert_allclose(ranks, expected)


def test_rank_distance_across(list_data, distance):
    """Test rank distance for across-category transitions."""
    ranks = stats.rank_distance(
        distance,
        list_data['pool_items'],
        list_data['recall_items'],
        list_data['pool_items'],
        list_data['recall_items'],
        pool_test=list_data['pool_test'],
        recall_test=list_data['recall_test'],
        test=lambda x, y: x != y,
    )
    expected = np.array([1 - (0.5 / 3), 1 - (0.5 / 1)])
    np.testing.assert_allclose(ranks, expected)


def test_rank_distance_shifted(list_data, distance):
    """Test rank distance based on shifted recalls."""
    list_data['recall_items'] = [[4, 3, 1, 7, 3, 0, 6, 2]]
    distance[3, 1] = 4
    distance[1, 3] = 4
    # sequences: [[4, 3, 1], [3, 1, 7], [0, 6, 5]]
    # -1
    # actual: [4, 2, 2]
    # possible: [
    #     [1, 1, 2, 2, 2, 4],
    #     [1, 1, 2, 2, 2],
    #     [2, 3],
    # ]
    # rank: [0.0, 0.25, 1.0]
    # -2
    # actual: [2, 2, 1]
    # possible: [
    #     [2, 2, 2, 3, 3, 3],
    #     [1, 1, 2, 2, 2],
    #     [1, 2],
    # ]
    # rank: [0.8, 0.25, 1.0]
    ranks = stats.rank_distance_shifted(
        distance,
        2,
        list_data['pool_items'],
        list_data['recall_items'],
        list_data['pool_items'],
        list_data['recall_items'],
    )
    expected = np.array([[0.8, 0.0], [0.25, 0.25], [1.0, 1.0]])
    np.testing.assert_allclose(ranks, expected)


def test_rank_distance_window():
    """Test rank distance relative to items in a window."""
    distance = np.array(
        [
            [0, 3, 1, 1, 2, 2, 2, 2],
            [3, 0, 1, 1, 2, 2, 2, 2],
            [1, 1, 0, 4, 2, 2, 2, 2],
            [1, 1, 4, 0, 2, 2, 2, 2],
            [2, 2, 2, 2, 0, 5, 3, 3],
            [2, 2, 2, 2, 5, 0, 3, 3],
            [2, 2, 2, 2, 3, 3, 0, 6],
            [2, 2, 2, 2, 3, 3, 6, 0],
        ]
    )
    pool = [[1, 2, 3, 4, 5, 6, 7, 8]]
    outputs = [[4, 2, 1, 7, 5, 3, 4]]
    pool_index = [[4, 5, 6, 7, 0, 1, 2, 3]]
    outputs_index = [[3, 1, 0, 6, 4, 2, 3]]
    list_length = 8
    window_lags = [-1, 0, 1]

    # included: [(4, 2), (7, 5), (5, 3)]
    # prev: [[3, 4, 5], [6, 7, 8], [4, 5, 6]]
    # curr: [2, 5, 3]
    # poss: [[1, 2, 6, 7, 8], [3, 5], [3, 8]]

    # prev: [[6, 7, 0], [1, 2, 3], [7, 0, 1]]
    # curr: [5, 0, 6]
    # poss: [[4, 5, 1, 2, 3], [6, 0], [6, 3]]

    # -1
    # actual: [3, 3, 6]
    # possible: [[3, 3, 2, 2, 2], [2, 3], [6, 2]]
    # rank: [0.125, 0.0, 0.0]

    # 0
    # actual: [3, 1, 2]
    # possible: [[3, 3, 2, 2, 2], [2, 1], [2, 1]]
    # rank: [0.125, 1.0, 0.0]

    # +1
    # actual: [2, 1, 2]
    # possible: [[2, 2, 3, 1, 1], [2, 1], [2, 1]]
    # rank: [0.375, 1.0, 0.0]
    ranks = stats.rank_distance_window(
        distance, list_length, window_lags, pool, outputs, pool_index, outputs_index
    )
    expected = np.array([[0.125, 0.125, 0.375], [0, 1, 1], [0, 0, 0]])
    np.testing.assert_allclose(ranks, expected)



"""Utilities for working with free recall data."""

from pkg_resources import resource_filename
import warnings
import numpy as np
import pandas as pd
import seaborn as sns

from psifr import stats
from psifr import measures


def sample_data(study):
    """Read sample data."""
    data_file = resource_filename('psifr', f'data/{study}.csv')
    df = pd.read_csv(data_file)
    return df


def sample_distances(study):
    """Read sample distances."""
    distance_file = resource_filename('psifr', f'distances/{study}.npz')
    f = np.load(distance_file)
    return f['items'], f['distances']


def table_from_lists(subjects, study, recall, lists=None, **kwargs):
    """
    Create table format data from list format data.

    Parameters
    ----------
    subjects : list of hashable
        Subject identifier for each list.

    study : list of list of hashable
        List of items for each study list.

    recall : list of list of hashable
        List of recalled items for each study list.

    lists : list of hashable, optional
        List of list numbers. If not specified, lists for each subject
        will be numbered sequentially starting from one.

    Returns
    -------
    data : pandas.DataFrame
        Data in table format.

    See Also
    --------
    split_lists : Split a table into list format.

    Examples
    --------
    >>> from psifr import fr
    >>> subjects_list = [1, 1, 2, 2]
    >>> study_lists = [['a', 'b'], ['c', 'd'], ['e', 'f'], ['g', 'h']]
    >>> recall_lists = [['b'], ['d', 'c'], ['f', 'e'], []]
    >>> fr.table_from_lists(subjects_list, study_lists, recall_lists)
        subject  list trial_type  position item
    0         1     1      study         1    a
    1         1     1      study         2    b
    2         1     1     recall         1    b
    3         1     2      study         1    c
    4         1     2      study         2    d
    5         1     2     recall         1    d
    6         1     2     recall         2    c
    7         2     1      study         1    e
    8         2     1      study         2    f
    9         2     1     recall         1    f
    10        2     1     recall         2    e
    11        2     2      study         1    g
    12        2     2      study         2    h

    >>> subjects_list = [1, 1]
    >>> study_lists = [['a', 'b'], ['c', 'd']]
    >>> recall_lists = [['b'], ['d', 'c']]
    >>> col1 = ([[1, 2], [1, 2]], [[2], [2, 1]])
    >>> col2 = ([[1, 1], [2, 2]], None)
    >>> fr.table_from_lists(subjects_list, study_lists, recall_lists, col1=col1, col2=col2)
       subject  list trial_type  position item  col1  col2
    0        1     1      study         1    a     1   1.0
    1        1     1      study         2    b     2   1.0
    2        1     1     recall         1    b     2   NaN
    3        1     2      study         1    c     1   2.0
    4        1     2      study         2    d     2   2.0
    5        1     2     recall         1    d     2   NaN
    6        1     2     recall         2    c     1   NaN
    """
    assert len(subjects) == len(study) == len(recall), 'Input lengths must match.'
    d = {'subject': [], 'list': [], 'trial_type': [], 'position': [], 'item': []}
    for key in kwargs.keys():
        d[key] = []
    prev_subject = None
    current_list = 1
    if lists is None:
        lists = [None] * len(subjects)
    else:
        assert len(subjects) == len(lists), 'Length of lists must match subjects.'
    labels = zip(subjects, study, recall, lists)
    for i, (subject, study_list, recall_list, n) in enumerate(labels):
        # set list number
        if n is not None:
            current_list = n
        elif subject != prev_subject:
            current_list = 1

        # add study events
        for j, study_item in enumerate(study_list):
            d['subject'].append(subject)
            d['list'].append(current_list)
            d['trial_type'].append('study')
            d['position'].append(j + 1)
            d['item'].append(study_item)
            for key, val in kwargs.items():
                if val[0] is not None:
                    d[key].append(val[0][i][j])
                else:
                    d[key].append(np.nan)

        # add recall events
        for j, recall_item in enumerate(recall_list):
            d['subject'].append(subject)
            d['list'].append(current_list)
            d['trial_type'].append('recall')
            d['position'].append(j + 1)
            d['item'].append(recall_item)
            for key, val in kwargs.items():
                if val[1] is not None:
                    d[key].append(val[1][i][j])
                else:
                    d[key].append(np.nan)
        current_list += 1
        prev_subject = subject
    data = pd.DataFrame(d)
    return data


def _match_values(series, values):
    """Get matches for a data column."""
    if not hasattr(values, '__iter__') or isinstance(values, str):
        values = [values]
    include = series.isin(values)
    return include


def filter_data(
    data,
    subjects=None,
    lists=None,
    trial_type=None,
    positions=None,
    inputs=None,
    outputs=None,
):
    """
    Filter data to get a subset of trials.

    Parameters
    ----------
    data : pandas.DataFrame
        Raw or merged data to filter.

    subjects : hashable or list of hashable
        Subject or subjects to include.

    lists : hashable or list of hashable
        List or lists to include.

    trial_type : {'study', 'recall'}
        Trial type to include.

    positions : int or list of int
        Position or positions to include.

    inputs : int or list of int
        Input position or positions to include.

    outputs : int or list of int
        Output position or positions to include.

    Returns
    -------
    filtered : pandas.DataFrame
        The filtered subset of data.

    Examples
    --------
    >>> from psifr import fr
    >>> subjects_list = [1, 1, 2, 2]
    >>> study_lists = [['a', 'b'], ['c', 'd'], ['e', 'f'], ['g', 'h']]
    >>> recall_lists = [['b'], ['d', 'c'], ['f', 'e'], []]
    >>> raw = fr.table_from_lists(subjects_list, study_lists, recall_lists)
    >>> fr.filter_data(raw, subjects=1, trial_type='study')
       subject  list trial_type  position item
    0        1     1      study         1    a
    1        1     1      study         2    b
    3        1     2      study         1    c
    4        1     2      study         2    d

    >>> data = fr.merge_free_recall(raw)
    >>> fr.filter_data(data, subjects=2)
       subject  list item  input  output  study  recall  repeat  intrusion  prior_list  prior_input
    4        2     1    e      1     2.0   True    True       0      False         NaN          NaN
    5        2     1    f      2     1.0   True    True       0      False         NaN          NaN
    6        2     2    g      1     NaN   True   False       0      False         NaN          NaN
    7        2     2    h      2     NaN   True   False       0      False         NaN          NaN
    """
    include = data['subject'].notna()
    if subjects is not None:
        include &= _match_values(data['subject'], subjects)

    if lists is not None:
        include &= _match_values(data['list'], lists)

    if trial_type is not None:
        include &= data['trial_type'] == trial_type

    if positions is not None:
        include &= _match_values(data['position'], positions)

    if inputs is not None:
        include &= _match_values(data['input'], inputs)

    if outputs is not None:
        include &= _match_values(data['output'], outputs)

    filtered = data.loc[include].copy()
    return filtered


def check_data(df):
    """Run checks on free recall data.

    Parameters
    ----------
    df : pandas.DataFrame
        Contains one row for each trial (study and recall). Must have fields:
            subject : number or str
                Subject identifier.

            list : number
                List identifier. This applies to both study and recall trials.

            trial_type : str
                Type of trial; may be 'study' or 'recall'.

            position : number
                Position within the study list or recall sequence.

            item : str
                Item that was either presented or recalled on this trial.

    Examples
    --------
    >>> from psifr import fr
    >>> import pandas as pd
    >>> raw = pd.DataFrame(
    ...     {'subject': [1, 1], 'list': [1, 1], 'position': [1, 2], 'item': ['a', 'b']}
    ... )
    >>> fr.check_data(raw)
    Traceback (most recent call last):
      File "psifr/fr.py", line 253, in check_data
        assert col in df.columns, f'Required column {col} is missing.'
    AssertionError: Required column trial_type is missing.
    """
    # check that all fields are accounted for
    columns = ['subject', 'list', 'trial_type', 'position', 'item']
    for col in columns:
        assert col in df.columns, f'Required column {col} is missing.'

    # only one column has a hard constraint on its exact content
    assert (
        df['trial_type'].isin(['study', 'recall']).all()
    ), 'trial_type for all trials must be "study" or "recall".'


def block_index(list_labels):
    """
    Get index of each block in a list.

    Parameters
    ----------
    list_labels : list or numpy.ndarray
        Position labels that define the blocks.

    Returns
    -------
    block : numpy.ndarray
        Block index of each position.

    Examples
    --------
    >>> from psifr import fr
    >>> list_labels = [2, 2, 3, 3, 3, 1, 1]
    >>> fr.block_index(list_labels)
    array([1, 1, 2, 2, 2, 3, 3])
    """
    prev_label = ''
    curr_block = 0
    block = np.zeros(len(list_labels), dtype=int)
    for i, label in enumerate(list_labels):
        if prev_label != label:
            curr_block += 1
        block[i] = curr_block
        prev_label = label
    return block


def pool_index(trial_items, pool_items_list):
    """
    Get the index of each item in the full pool.

    Parameters
    ----------
    trial_items : pandas.Series
        The item presented on each trial.

    pool_items_list : list or numpy.ndarray
        List of items in the full pool.

    Returns
    -------
    item_index : pandas.Series
        Index of each item in the pool. Trials with items not in the
        pool will be <NA>.

    Examples
    --------
    >>> import pandas as pd
    >>> from psifr import fr
    >>> trial_items = pd.Series(['b', 'a', 'z', 'c', 'd'])
    >>> pool_items_list = ['a', 'b', 'c', 'd', 'e', 'f']
    >>> fr.pool_index(trial_items, pool_items_list)
    0       1
    1       0
    2    <NA>
    3       2
    4       3
    dtype: Int64
    """
    pool_map = dict(zip(pool_items_list, np.arange(len(pool_items_list))))
    item_index = trial_items.map(pool_map).astype('Int64')
    return item_index


def reset_list(df):
    """
    Reset list index in a DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        Raw or merged data. Must have subject and list fields.

    Returns
    -------
    pandas.DataFrame
        Data with a renumbered list field, starting from 1.

    Examples
    --------
    >>> from psifr import fr
    >>> subjects_list = [1, 1]
    >>> study_lists = [['a', 'b'], ['c', 'd']]
    >>> recall_lists = [['b'], ['c', 'd']]
    >>> list_nos = [3, 4]
    >>> raw = fr.table_from_lists(subjects_list, study_lists, recall_lists, lists=list_nos)
    >>> raw
       subject  list trial_type  position item
    0        1     3      study         1    a
    1        1     3      study         2    b
    2        1     3     recall         1    b
    3        1     4      study         1    c
    4        1     4      study         2    d
    5        1     4     recall         1    c
    6        1     4     recall         2    d

    >>> fr.reset_list(raw)
       subject  list trial_type  position item
    0        1     1      study         1    a
    1        1     1      study         2    b
    2        1     1     recall         1    b
    3        1     2      study         1    c
    4        1     2      study         2    d
    5        1     2     recall         1    c
    6        1     2     recall         2    d
    """
    df = df.copy()
    for subject in df['subject'].unique():
        subject_lists = df.loc[df['subject'] == subject, 'list'].unique()
        for i, listno in enumerate(subject_lists):
            df.loc[(df['subject'] == subject) & (df['list'] == listno), 'list'] = i + 1
    return df


def split_lists(frame, phase, keys=None, names=None, item_query=None, as_list=False):
    """
    Convert free recall data from one phase to split format.

    Parameters
    ----------
    frame : pandas.DataFrame
        Free recall data with separate study and recall events.

    phase : {'study', 'recall', 'raw'}
        Phase of recall to split. If 'raw', all trials will be included.

    keys : list of str, optional
        Data columns to include in the split data. If not specified,
        all columns will be included.

    names : list of str, optional
        Name for each column in the returned split data. Default is to
        use the same names as the input columns.

    item_query : str, optional
        Query string to select study trials to include. See
        `pandas.DataFrame.query` for allowed format.

    as_list : bool, optional
        If true, each column will be output as a list; otherwise,
        outputs will be numpy.ndarray.

    Returns
    -------
    split : dict of str: list
        Data in split format. Each included column will be a key in the
        dictionary, with a list of either numpy.ndarray (default) or
        lists, containing the values for that column.

    See Also
    --------
    table_from_lists : Convert list-format data to a table.

    Examples
    --------
    >>> from psifr import fr
    >>> study = [['absence', 'hollow'], ['fountain', 'piano']]
    >>> recall = [['absence'], ['piano', 'fountain']]
    >>> raw = fr.table_from_lists([1, 1], study, recall)
    >>> data = fr.merge_free_recall(raw)
    >>> data
       subject  list      item  input  output  study  recall  repeat  intrusion  prior_list  prior_input
    0        1     1   absence      1     1.0   True    True       0      False         NaN          NaN
    1        1     1    hollow      2     NaN   True   False       0      False         NaN          NaN
    2        1     2  fountain      1     2.0   True    True       0      False         NaN          NaN
    3        1     2     piano      2     1.0   True    True       0      False         NaN          NaN

    Get study events split by list, just including the list and item fields.

    >>> fr.split_lists(data, 'study', keys=['list', 'item'], as_list=True)
    {'list': [[1, 1], [2, 2]], 'item': [['absence', 'hollow'], ['fountain', 'piano']]}

    Export recall events, split by list.

    >>> fr.split_lists(data, 'recall', keys=['item'], as_list=True)
    {'item': [['absence'], ['piano', 'fountain']]}

    Raw events (i.e., events that haven't been scored) can also be
    exported to list format.

    >>> fr.split_lists(raw, 'raw', keys=['position'])
    {'position': [array([1, 2, 1]), array([1, 2, 1, 2])]}
    """
    split = {}
    if keys is None:
        keys = frame.columns.tolist()
    if not keys:
        return split
    if names is None:
        names = keys

    unique_lists = frame['list'].unique()
    if phase == 'study':
        phase_data = frame.loc[frame['study']]
    elif phase == 'recall':
        phase_data = frame.loc[frame['recall']].sort_values(['list', 'output'])
    elif phase == 'raw':
        phase_data = frame
    else:
        raise ValueError(f'Invalid phase: {phase}')

    if item_query is not None:
        # get the subset of the pool that is of interest
        mask = phase_data.eval(item_query).to_numpy()
    else:
        mask = np.ones(phase_data.shape[0], dtype=bool)

    frame_idx = phase_data.reset_index().groupby('list').indices
    for key, name in zip(keys, names):
        if key is None or key not in frame.columns:
            split[name] = None
            continue
        all_values = phase_data[key].to_numpy()
        split[name] = []
        for n in unique_lists:
            if n in frame_idx:
                list_idx = frame_idx[n]
                x = all_values[list_idx][mask[list_idx]]
            else:
                x = np.array([])

            if as_list:
                split[name].append(x.tolist())
            else:
                split[name].append(x)
    return split


def merge_free_recall(data, **kwargs):
    """
    Score free recall data by matching up study and recall events.

    Parameters
    ----------
    data : pandas.DataFrame
        Free recall data in Psifr format. Must have subject, list,
        trial_type, position, and item columns.

    merge_keys : list, optional
        Columns to use to designate events to merge. Default is
        ['subject', 'list', 'item'], which will merge events related to
        the same item, but only within list.

    list_keys : list, optional
        Columns that apply to both study and recall events.

    study_keys : list, optional
        Columns that only apply to study events.

    recall_keys : list, optional
        Columns that only apply to recall events.

    position_key : str, optional
        Column indicating the position of each item in either the study
        list or the recall sequence.

    Returns
    -------
    merged : pandas.DataFrame
        Merged information about study and recall events. Each row
        corresponds to one unique input/output pair.

        The following columns will be added:

        input : int
            Position of each item in the input list (i.e., serial
            position).

        output : int
            Position of each item in the recall sequence.

        study : bool
            True for rows corresponding to a unique study event.

        recall : bool
            True for rows corresponding to a unique recall event.

        repeat : int
            Number of times this recall event has been repeated (0 for
            the first recall of an item).

        intrusion : bool
            True for recalls that do not correspond to any study event.

        prior_list : int
            For prior-list intrusions, the list the item was presented.

        prior_position : int
            For prior-list intrusions, the position the item was presented.

    See Also
    --------
    merge_lists : Flexibly merge study events with recall events.
        Useful for recall phases that don't match the typical free
        recall setup, like final free recall of all lists.

    Examples
    --------
    >>> import numpy as np
    >>> from psifr import fr
    >>> study = [['absence', 'hollow'], ['fountain', 'piano']]
    >>> recall = [['absence'], ['piano', 'hollow']]
    >>> raw = fr.table_from_lists([1, 1], study, recall)
    >>> raw
       subject  list trial_type  position      item
    0        1     1      study         1   absence
    1        1     1      study         2    hollow
    2        1     1     recall         1   absence
    3        1     2      study         1  fountain
    4        1     2      study         2     piano
    5        1     2     recall         1     piano
    6        1     2     recall         2    hollow

    Score the data to create a table with matched study and recall events.

    >>> data = fr.merge_free_recall(raw)
    >>> data
       subject  list      item  input  output  study  recall  repeat  intrusion  prior_list  prior_input
    0        1     1   absence    1.0     1.0   True    True       0      False         NaN          NaN
    1        1     1    hollow    2.0     NaN   True   False       0      False         NaN          NaN
    2        1     2  fountain    1.0     NaN   True   False       0      False         NaN          NaN
    3        1     2     piano    2.0     1.0   True    True       0      False         NaN          NaN
    4        1     2    hollow    NaN     2.0  False    True       0       True         1.0          2.0

    You can also include non-standard columns. Information that only
    applies to study events (here, the encoding task used) can be
    indicated using the :code:`study_keys` input.

    >>> raw['task'] = np.array([1, 2, np.nan, 2, 1, np.nan, np.nan])
    >>> fr.merge_free_recall(raw, study_keys=['task'])
       subject  list      item  input  output  study  recall  repeat  intrusion  task  prior_list  prior_input
    0        1     1   absence    1.0     1.0   True    True       0      False   1.0         NaN          NaN
    1        1     1    hollow    2.0     NaN   True   False       0      False   2.0         NaN          NaN
    2        1     2  fountain    1.0     NaN   True   False       0      False   2.0         NaN          NaN
    3        1     2     piano    2.0     1.0   True    True       0      False   1.0         NaN          NaN
    4        1     2    hollow    NaN     2.0  False    True       0       True   NaN         1.0          2.0

    Information that only applies to recall onsets (here, the time in
    seconds after the start of the recall phase that a recall attempt
    was made), can be indicated using the :code:`recall_keys` input.

    >>> raw['onset'] = np.array([np.nan, np.nan, 1.1, np.nan, np.nan, 1.4, 3.8])
    >>> fr.merge_free_recall(raw, recall_keys=['onset'])
       subject  list      item  input  output  study  recall  repeat  intrusion  onset  prior_list  prior_input
    0        1     1   absence    1.0     1.0   True    True       0      False    1.1         NaN          NaN
    1        1     1    hollow    2.0     NaN   True   False       0      False    NaN         NaN          NaN
    2        1     2  fountain    1.0     NaN   True   False       0      False    NaN         NaN          NaN
    3        1     2     piano    2.0     1.0   True    True       0      False    1.4         NaN          NaN
    4        1     2    hollow    NaN     2.0  False    True       0       True    3.8         1.0          2.0

    Use :code:`list_keys` to indicate columns that apply to both study
    and recall events. If :code:`list_keys` do not match for a pair of
    study and recall events, they will not be matched in the output.

    >>> raw['condition'] = np.array([1, 1, 1, 2, 2, 2, 2])
    >>> fr.merge_free_recall(raw, list_keys=['condition'])
       subject  list      item  input  output  study  recall  repeat  intrusion  condition  prior_list  prior_input
    0        1     1   absence    1.0     1.0   True    True       0      False          1         NaN          NaN
    1        1     1    hollow    2.0     NaN   True   False       0      False          1         NaN          NaN
    2        1     2  fountain    1.0     NaN   True   False       0      False          2         NaN          NaN
    3        1     2     piano    2.0     1.0   True    True       0      False          2         NaN          NaN
    4        1     2    hollow    NaN     2.0  False    True       0       True          2         1.0          2.0
    """
    study = data.loc[data['trial_type'] == 'study'].copy()
    recall = data.loc[data['trial_type'] == 'recall'].copy()
    merged = merge_lists(study, recall, **kwargs)

    # to identify prior-list intrusions, merge study events and intrusions
    intrusions = merged.query('intrusion')
    plis = pd.merge(intrusions, study, on=['subject', 'item'], how='inner')

    # add prior list and prior input information
    plis['list'] = plis['list_x']
    plis['prior_list'] = plis['list_y']
    plis['prior_input'] = plis['position']
    include = ['subject', 'list', 'item', 'output', 'prior_list', 'prior_input']
    merged = pd.merge(
        merged, plis[include], on=['subject', 'list', 'item', 'output'], how='outer'
    )

    # reset concidental "future list intrusions"
    isfli = merged['list'] < merged['prior_list']
    merged.loc[isfli, 'prior_list'] = np.nan
    merged.loc[isfli, 'prior_input'] = np.nan
    return merged


def merge_lists(
    study,
    recall,
    merge_keys=None,
    list_keys=None,
    study_keys=None,
    recall_keys=None,
    position_key='position',
):
    """
    Merge study and recall events together for each list.

    Parameters
    ----------
    study : pandas.DataFrame
        Information about all study events. Should have one row for
        each study event.

    recall : pandas.DataFrame
        Information about all recall events. Should have one row for
        each recall attempt.

    merge_keys : list, optional
        Columns to use to designate events to merge. Default is
        ['subject', 'list', 'item'], which will merge events related to
        the same item, but only within list.

    list_keys : list, optional
        Columns that apply to both study and recall events.

    study_keys : list, optional
        Columns that only apply to study events.

    recall_keys : list, optional
        Columns that only apply to recall events.

    position_key : str, optional
        Column indicating the position of each item in either the study
        list or the recall sequence.

    Returns
    -------
    merged : pandas.DataFrame
        Merged information about study and recall events. Each row
        corresponds to one unique input/output pair.

        The following columns will be added:

        input : int
            Position of each item in the input list (i.e., serial
            position).

        output : int
            Position of each item in the recall sequence.

        study : bool
            True for rows corresponding to a unique study event.

        recall : bool
            True for rows corresponding to a unique recall event.

        repeat : int
            Number of times this recall event has been repeated (0 for
            the first recall of an item).

        intrusion : bool
            True for recalls that do not correspond to any study event.

    See Also
    --------
    merge_free_recall : Score standard free recall data.

    Examples
    --------
    >>> import pandas as pd
    >>> from psifr import fr
    >>> study = pd.DataFrame(
    ...    {'subject': [1, 1], 'list': [1, 1], 'position': [1, 2], 'item': ['a', 'b']}
    ... )
    >>> recall = pd.DataFrame(
    ...    {'subject': [1], 'list': [1], 'position': [1], 'item': ['b']}
    ... )
    >>> fr.merge_lists(study, recall)
       subject  list item  input  output  study  recall  repeat  intrusion
    0        1     1    a      1     NaN   True   False       0      False
    1        1     1    b      2     1.0   True    True       0      False
    """
    if merge_keys is None:
        merge_keys = ['subject', 'list', 'item']

    if list_keys is None:
        list_keys = []

    if study_keys is None:
        study_keys = []

    if recall_keys is None:
        recall_keys = []

    # get running count of number of times each item is recalled in each list
    recall = recall.copy()
    recall.loc[:, 'repeat'] = recall.groupby(merge_keys).cumcount()

    # get just the fields to use in the merge
    study = study.copy()
    study = study[merge_keys + ['position'] + list_keys + study_keys]
    recall = recall[merge_keys + ['repeat', 'position'] + list_keys + recall_keys]

    # merge information from study and recall trials
    merged = pd.merge(
        study,
        recall,
        left_on=merge_keys + list_keys,
        right_on=merge_keys + list_keys,
        how='outer',
    )

    # position from study events indicates input position;
    # position from recall events indicates output position
    merged = merged.rename(
        columns={position_key + '_x': 'input', position_key + '_y': 'output'}
    )

    # fix repeats field to define for non-recalled items
    merged.loc[merged['repeat'].isna(), 'repeat'] = 0
    merged = merged.astype({'repeat': 'int'})

    # field to indicate unique study events
    merged.loc[:, 'study'] = merged['input'].notna() & (merged['repeat'] == 0)

    # TODO: deal with repeats in the study list
    # field to indicate unique recall events
    merged.loc[:, 'recall'] = merged['output'].notna()

    # field to indicate whether a given recall was an intrusion
    merged.loc[:, 'intrusion'] = merged['input'].isna()

    # reorder columns
    core_keys = ['input', 'output', 'study', 'recall', 'repeat', 'intrusion']
    columns = merge_keys + core_keys + list_keys + study_keys + recall_keys
    merged = merged.reindex(columns=columns)

    # sort rows in standard order
    sort_keys = merge_keys.copy() + ['input']
    sort_keys.remove('item')
    merged = merged.sort_values(by=sort_keys, ignore_index=True)

    return merged


def spc(df):
    """
    Serial position curve.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists.

    Returns
    -------
    recall : pandas.DataFrame
        subject : hashable
            Subject identifier.

        input : int
            Serial position in the list.

        recall : float
            Recall probability for each serial position.

    See Also
    --------
    plot_spc : Plot serial position curve results.
    pnr : Probability of nth recall.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> fr.spc(data)
         subject  input    recall
    0          1    1.0  0.541667
    1          1    2.0  0.458333
    2          1    3.0  0.625000
    3          1    4.0  0.333333
    4          1    5.0  0.437500
    ..       ...    ...       ...
    955       47   20.0  0.500000
    956       47   21.0  0.770833
    957       47   22.0  0.729167
    958       47   23.0  0.895833
    959       47   24.0  0.958333
    <BLANKLINE>
    [960 rows x 3 columns]
    """
    clean = df.query('study')
    recall = clean.groupby(['subject', 'input'])['recall'].mean()
    recall = pd.DataFrame(recall).reset_index()
    return recall


def pnr(df, item_query=None, test_key=None, test=None):
    """
    Probability of recall by serial position and output position.

    Calculate probability of Nth recall, where N is each output
    position. Invalid recalls (repeats and intrusions) are ignored and
    not counted toward output position.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, study, recall.
        Input position must be defined such that the first serial
        position is 1, not 0.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    prob : pandas.DataFrame
        Analysis results. Has fields: subject, output, input, prob,
        actual, possible. The prob column for output x and input y
        indicates the probability of recalling input position y at
        output position x. The actual and possible columns give the
        raw tallies for how many times an event actually occurred and
        how many times it was possible given the recall sequence.

    See Also
    --------
    plot_spc : Plot recall probability as a function of serial
        position.
    spc : Overall recall probability by serial position.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> fr.pnr(data)
           subject  output  input      prob  actual  possible
    0            1       1      1  0.000000       0        48
    1            1       1      2  0.020833       1        48
    2            1       1      3  0.000000       0        48
    3            1       1      4  0.000000       0        48
    4            1       1      5  0.000000       0        48
    ...        ...     ...    ...       ...     ...       ...
    23035       47      24     20       NaN       0         0
    23036       47      24     21       NaN       0         0
    23037       47      24     22       NaN       0         0
    23038       47      24     23       NaN       0         0
    23039       47      24     24       NaN       0         0
    <BLANKLINE>
    [23040 rows x 6 columns]
    """
    list_length = int(df['input'].max())
    measure = measures.TransitionOutputs(
        list_length, item_query=item_query, test_key=test_key, test=test
    )
    prob = measure.analyze(df)
    return prob


def _subject_pli_list_lag(df, max_lag):
    """List lag of prior-list intrusions for one subject."""
    if max_lag >= df['list'].nunique():
        warnings.warn('All lists are excluded based on max_lag.')
    results = pd.DataFrame(
        index=pd.Index(np.arange(1, max_lag + 1), name='list_lag'),
        columns=['count', 'per_list', 'prob'],
        dtype=float,
    )
    included = df[df['list'] > max_lag]
    intrusions = included.query('intrusion')
    if len(intrusions) > 0:
        list_lag = intrusions['list'] - intrusions['prior_list']
        results['count'] = list_lag.value_counts()
    results['count'] = results['count'].fillna(0).astype(int)
    results['per_list'] = results['count'] / included['list'].nunique()
    results['prob'] = results['count'] / intrusions.shape[0]
    return results


def pli_list_lag(df, max_lag):
    """
    List lag of prior-list intrusions.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_free_recall. Must have
        fields: subject, list, intrusion, prior_list. Lists must be
        numbered starting from 1 and all lists must be included.

    max_lag : int
        Maximum list lag to consider. The intial :code:`max_lag` lists
        for each subject will be excluded so that all considered lags
        are possible for all included lists.

    Returns
    -------
    results : pandas.DataFrame
        For each subject and list lag, the proportion of intrusions at
        that lag, in the :code:`results['prob']` column.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> fr.pli_list_lag(data, 3)
         subject  list_lag  count  per_list      prob
    0          1         1      7  0.155556  0.259259
    1          1         2      5  0.111111  0.185185
    2          1         3      0  0.000000  0.000000
    3          2         1      9  0.200000  0.191489
    4          2         2      2  0.044444  0.042553
    ..       ...       ...    ...       ...       ...
    115       46         2      1  0.022222  0.100000
    116       46         3      0  0.000000  0.000000
    117       47         1      5  0.111111  0.277778
    118       47         2      1  0.022222  0.055556
    119       47         3      0  0.000000  0.000000
    <BLANKLINE>
    [120 rows x 5 columns]
    """
    result = df.groupby('subject').apply(_subject_pli_list_lag, max_lag)
    result = result.reset_index()
    return result


def lag_crp(
    df, lag_key='input', count_unique=False, item_query=None, test_key=None, test=None
):
    """
    Lag-CRP for multiple subjects.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists. Must have fields:
        subject, list, input, output, recalled. Input position must be
        defined such that the first serial position is 1, not 0.

    lag_key : str, optional
        Name of column to use when calculating lag between recalled
        items. Default is to calculate lag based on input position.

    count_unique : bool, optional
        If true, possible transitions of the same lag will only be
        incremented once per transition.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    results : pandas.DataFrame
        Has fields:

        subject : hashable
            Results are separated by each subject.

        lag : int
            Lag of input position between two adjacent recalls.

        prob : float
            Probability of each lag transition.

        actual : int
            Total of actual made transitions at each lag.

        possible : int
            Total of times each lag was possible, given the prior
            input position and the remaining items to be recalled.

    See Also
    --------
    lag_rank : Rank of the absolute lags in recall sequences.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> fr.lag_crp(data)
          subject   lag      prob  actual  possible
    0           1 -23.0  0.020833       1        48
    1           1 -22.0  0.035714       3        84
    2           1 -21.0  0.026316       3       114
    3           1 -20.0  0.024000       3       125
    4           1 -19.0  0.014388       2       139
    ...       ...   ...       ...     ...       ...
    1875       47  19.0  0.061224       3        49
    1876       47  20.0  0.055556       2        36
    1877       47  21.0  0.045455       1        22
    1878       47  22.0  0.071429       1        14
    1879       47  23.0  0.000000       0         6
    <BLANKLINE>
    [1880 rows x 5 columns]
    """
    list_length = df[lag_key].max()
    measure = measures.TransitionLag(
        list_length,
        lag_key=lag_key,
        count_unique=count_unique,
        item_query=item_query,
        test_key=test_key,
        test=test,
    )
    crp = measure.analyze(df)
    return crp


def lag_crp_compound(
    df, lag_key='input', count_unique=False, item_query=None, test_key=None, test=None
):
    """
    Conditional response probability by lag of current and prior transitions.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists. Must have fields:
        subject, list, input, output, recalled. Input position must be
        defined such that the first serial position is 1, not 0.

    lag_key : str, optional
        Name of column to use when calculating lag between recalled
        items. Default is to calculate lag based on input position.

    count_unique : bool, optional
        If true, possible transitions of the same lag will only be
        incremented once per transition.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    results : pandas.DataFrame
        Has fields:

        subject : hashable
            Results are separated by each subject.

        previous : int
            Lag of the previous transition.

        current : int
            Lag of the current transition.

        prob : float
            Probability of each lag transition.

        actual : int
            Total of actual made transitions at each lag.

        possible : int
            Total of times each lag was possible, given the prior
            input position and the remaining items to be recalled.

    See Also
    --------
    lag_crp : Conditional response probability by lag.

    Examples
    --------
    >>> from psifr import fr
    >>> subjects = [1]
    >>> study = [['absence', 'hollow', 'pupil', 'fountain']]
    >>> recall = [['fountain', 'hollow', 'absence']]
    >>> raw = fr.table_from_lists(subjects, study, recall)
    >>> data = fr.merge_free_recall(raw)
    >>> crp = fr.lag_crp_compound(data)
    >>> crp.head(14)
        subject  previous  current  prob  actual  possible
    0         1        -3       -3   NaN       0         0
    1         1        -3       -2   NaN       0         0
    2         1        -3       -1   NaN       0         0
    3         1        -3        0   NaN       0         0
    4         1        -3        1   NaN       0         0
    5         1        -3        2   NaN       0         0
    6         1        -3        3   NaN       0         0
    7         1        -2       -3   NaN       0         0
    8         1        -2       -2   NaN       0         0
    9         1        -2       -1   1.0       1         1
    10        1        -2        0   NaN       0         0
    11        1        -2        1   0.0       0         1
    12        1        -2        2   NaN       0         0
    13        1        -2        3   NaN       0         0
    """
    list_length = df[lag_key].max()
    measure = measures.TransitionLag(
        list_length,
        lag_key=lag_key,
        count_unique=count_unique,
        item_query=item_query,
        test_key=test_key,
        test=test,
        compound=True,
    )
    crp = measure.analyze(df)
    return crp


def lag_rank(df, lag_key='input', item_query=None, test_key=None, test=None):
    """
    Calculate rank of the absolute lags in free recall lists.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, recalled.
        Input position must be defined such that the first serial
        position is 1, not 0.

    lag_key : str, optional
        Name of column to use when calculating lag between recalled
        items. Default is to calculate lag based on input position.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    stat : pandas.DataFrame
        Has fields 'subject' and 'rank'.

    See Also
    --------
    lag_crp : Conditional response probability by input lag.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> lag_rank = fr.lag_rank(data)
    >>> lag_rank.head()
       subject      rank
    0        1  0.610953
    1        2  0.635676
    2        3  0.612607
    3        4  0.667090
    4        5  0.643923
    """
    measure = measures.TransitionLagRank(
        lag_key=lag_key, item_query=item_query, test_key=test_key, test=test
    )
    rank = measure.analyze(df)
    return rank


def distance_crp(
    df,
    index_key,
    distances,
    edges,
    centers=None,
    count_unique=False,
    item_query=None,
    test_key=None,
    test=None,
    drop_bin=False,
):
    """
    Conditional response probability by distance bin.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged free recall data.

    index_key : str
        Name of column containing the index of each item in the
        `distances` matrix.

    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    edges : array-like
        Edges of bins to apply to the distances.

    centers : array-like, optional
        Centers to label each bin with. If not specified, the center
        point between edges will be used.

    count_unique : bool, optional
        If true, possible transitions to a given distance bin will only
        count once for a given transition.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    crp : pandas.DataFrame
        Has fields:

        subject : hashable
            Results are separated by each subject.

        bin : int
            Distance bin.

        prob : float
            Probability of each distance bin.

        actual : int
            Total of actual transitions for each distance bin.

        possible : int
            Total of times each distance bin was possible, given the
            prior input position and the remaining items to be
            recalled.

    See Also
    --------
    pool_index : Given a list of presented items and an item pool, look
        up the pool index of each item.
    distance_rank : Calculate rank of transition distances.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.spatial.distance import squareform
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> items, distances = fr.sample_distances('Morton2013')
    >>> data['item_index'] = fr.pool_index(data['item'], items)
    >>> edges = np.percentile(squareform(distances), np.linspace(1, 99, 10))
    >>> fr.distance_crp(data, 'item_index', distances, edges)
         subject    center             bin      prob  actual  possible
    0          1  0.467532  (0.352, 0.583]  0.085456     151      1767
    1          1  0.617748  (0.583, 0.653]  0.067916      87      1281
    2          1  0.673656  (0.653, 0.695]  0.062500      65      1040
    3          1  0.711075  (0.695, 0.727]  0.051836      48       926
    4          1  0.742069  (0.727, 0.757]  0.050633      44       869
    ..       ...       ...             ...       ...     ...       ...
    355       47  0.742069  (0.727, 0.757]  0.062822      61       971
    356       47  0.770867  (0.757, 0.785]  0.030682      27       880
    357       47  0.800404  (0.785, 0.816]  0.040749      37       908
    358       47  0.834473  (0.816, 0.853]  0.046651      39       836
    359       47  0.897275  (0.853, 0.941]  0.028868      25       866
    <BLANKLINE>
    [360 rows x 6 columns]
    """
    measure = measures.TransitionDistance(
        index_key,
        distances,
        edges,
        centers=centers,
        count_unique=count_unique,
        item_query=item_query,
        test_key=test_key,
        test=test,
    )
    crp = measure.analyze(df)
    if drop_bin:
        crp = crp.drop(columns=['bin'])
    return crp


def distance_rank(df, index_key, distances, item_query=None, test_key=None, test=None):
    """
    Calculate rank of transition distances in free recall lists.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, recalled.
        Input position must be defined such that the first serial
        position is 1, not 0.

    index_key : str
        Name of column containing the index of each item in the
        `distances` matrix.

    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    stat : pandas.DataFrame
        Has fields 'subject' and 'rank'.

    See Also
    --------
    pool_index : Given a list of presented items and an item pool, look
        up the pool index of each item.
    distance_crp : Conditional response probability by distance bin.

    Examples
    --------
    >>> from scipy.spatial.distance import squareform
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> items, distances = fr.sample_distances('Morton2013')
    >>> data['item_index'] = fr.pool_index(data['item'], items)
    >>> dist_rank = fr.distance_rank(data, 'item_index', distances)
    >>> dist_rank.head()
       subject      rank
    0        1  0.635571
    1        2  0.571457
    2        3  0.627282
    3        4  0.637596
    4        5  0.646181
    """
    measure = measures.TransitionDistanceRank(
        index_key, distances, item_query=item_query, test_key=test_key, test=test
    )
    rank = measure.analyze(df)
    return rank


def distance_rank_shifted(
    df, index_key, distances, max_shift, item_query=None, test_key=None, test=None
):
    """
    Rank of transition distances relative to earlier items.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, recalled.
        Input position must be defined such that the first serial
        position is 1, not 0.

    index_key : str
        Name of column containing the index of each item in the
        `distances` matrix.

    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    max_shift : int
        Maximum number of items back for which to rank distances.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    stat : pandas.DataFrame
        Has fields 'subject' and 'rank'.

    See Also
    --------
    pool_index : Given a list of presented items and an item pool, look
        up the pool index of each item.
    distance_rank : Rank of transition distances relative to the
        just-previous item.

    Examples
    --------
    >>> from scipy.spatial.distance import squareform
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw)
    >>> items, distances = fr.sample_distances('Morton2013')
    >>> data['item_index'] = fr.pool_index(data['item'], items)
    >>> dist_rank = fr.distance_rank_shifted(data, 'item_index', distances, 3)
    >>> dist_rank
         subject  shift      rank
    0          1     -3  0.523426
    1          1     -2  0.559199
    2          1     -1  0.634392
    3          2     -3  0.475931
    4          2     -2  0.507574
    ..       ...    ...       ...
    115       46     -2  0.515332
    116       46     -1  0.603304
    117       47     -3  0.542951
    118       47     -2  0.565001
    119       47     -1  0.635415
    <BLANKLINE>
    [120 rows x 3 columns]
    """
    measure = measures.TransitionDistanceRankShifted(
        index_key,
        distances,
        max_shift,
        item_query=item_query,
        test_key=test_key,
        test=test,
    )
    rank = measure.analyze(df)
    return rank


def distance_rank_window(
    df, index_key, distances, window_lags, item_query=None, test_key=None, test=None
):
    """
    Rank of transition distances relative to items in a window.

    Transitions are ranked based on their distance relative to items
    at specified lags from the previous item in the input list.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, recalled.
        Input position must be defined such that the first serial
        position is 1, not 0.

    index_key : str
        Name of column containing the index of each item in the
        `distances` matrix.

    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    window_lags : array_like
        Serial position lags to include in the window.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    stat : pandas.DataFrame
        Has fields 'subject', 'lag', and 'rank'.
    """
    list_length = int(df['input'].max())
    measure = measures.TransitionDistanceRankWindow(
        index_key,
        distances,
        list_length,
        window_lags,
        item_query=item_query,
        test_key=test_key,
        test=test,
    )
    rank = measure.analyze(df)
    return rank


def category_crp(df, category_key, item_query=None, test_key=None, test=None):
    """
    Conditional response probability of within-category transitions.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_lists. List length is
        assumed to be the same for all lists within each subject.
        Must have fields: subject, list, input, output, recalled.

    category_key : str
        Name of column with category labels.

    item_query : str, optional
        Query string to select items to include in the pool of possible
        recalls to be examined. See `pandas.DataFrame.query` for
        allowed format.

    test_key : str, optional
        Name of column with labels to use when testing transitions for
        inclusion.

    test : callable, optional
        Callable that takes in previous and current item values and
        returns True for transitions that should be included.

    Returns
    -------
    results : pandas.DataFrame
        Has fields:

        subject : hashable
            Results are separated by each subject.

        prob : float
            Probability of each lag transition.

        actual : int
            Total of actual made transitions at each lag.

        possible : int
            Total of times each lag was possible, given the prior
            input position and the remaining items to be recalled.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> data = fr.merge_free_recall(raw, study_keys=['category'])
    >>> cat_crp = fr.category_crp(data, 'category')
    >>> cat_crp.head()
       subject      prob  actual  possible
    0        1  0.801147     419       523
    1        2  0.733456     399       544
    2        3  0.763158     377       494
    3        4  0.814882     449       551
    4        5  0.877273     579       660
    """
    measure = measures.TransitionCategory(
        category_key, item_query=item_query, test_key=test_key, test=test
    )
    crp = measure.analyze(df)
    return crp


def _subject_category_clustering(df, category_key):
    """Subject category clustering."""
    study = split_lists(df, 'study', keys=[category_key])
    recall = split_lists(df, 'recall', keys=[category_key])
    lbc = stats.lbc(study[category_key], recall[category_key])
    arc = stats.arc(recall[category_key])
    clust = pd.Series({'lbc': np.nanmean(lbc), 'arc': np.nanmean(arc)})
    return clust


def category_clustering(df, category_key):
    """
    Category clustering of recall sequences.

    Calculates ARC (adjusted ratio of clustering) and LBC (list-based
    clustering) statistics indexing recall clustering by category.

    The papers introducing these measures do not describe how to handle
    repeats and intrusions. Here, to maintain the assumptions of the
    measures, they are removed from the recall sequences.

    Note that ARC is undefined when only one category is recalled.
    Lists with undefined statistics will be excluded from calculation
    of mean subject-level statistics. To calculate for each list
    separately, group by list before calling the function. For example:
    :code:`df.groupby('list').apply(fr.category_clustering, 'category')`.

    Parameters
    ----------
    df : pandas.DataFrame
        Merged study and recall data. See merge_free_recall. Must have
        a field indicating the category of each study and recall event.

    category_key : str
        Column with category labels. Labels may be any hashable (e.g.,
        a str or int).

    Returns
    -------
    stats : pandas.DataFrame
        For each subject, includes columns with the mean ARC and LBC
        statistics.

    Examples
    --------
    >>> from psifr import fr
    >>> raw = fr.sample_data('Morton2013')
    >>> mixed = raw.query('list_type == "mixed"')
    >>> data = fr.merge_free_recall(mixed, list_keys=['category'])
    >>> stats = fr.category_clustering(data, 'category')
    >>> stats.head()
       subject       lbc       arc
    0        1  3.657971  0.614545
    1        2  2.953623  0.407839
    2        3  3.363768  0.627371
    3        4  4.444928  0.688761
    4        5  7.530435  0.873755
    """
    # these analyses are undefined when there are repeats and
    # intrusions, so strip them out
    clean = df.query('~intrusion and repeat == 0')
    stats = clean.groupby('subject').apply(_subject_category_clustering, category_key)
    stats = stats.reset_index()
    return stats


def plot_spc(recall, **facet_kws):
    """
    Plot a serial position curve.

    Additional arguments are passed to seaborn.relplot.

    Parameters
    ----------
    recall : pandas.DataFrame
        Results from calling `spc`.
    """
    y = 'recall' if 'recall' in recall else 'prob'
    g = sns.FacetGrid(dropna=False, **facet_kws, data=recall.reset_index())
    g.map_dataframe(sns.lineplot, x='input', y=y)
    g.set_xlabels('Serial position')
    g.set_ylabels('Recall probability')
    g.set(ylim=(0, 1))
    return g


def plot_lag_crp(recall, max_lag=5, lag_key='lag', split=True, **facet_kws):
    """
    Plot conditional response probability by lag.

    Additional arguments are passed to seaborn.FacetGrid.

    Parameters
    ----------
    recall : pandas.DataFrame
        Results from calling `lag_crp`.

    max_lag : int, optional
        Maximum absolute lag to plot.

    lag_key : str, optional
        Name of the column indicating lag.

    split : bool, optional
        If true, will plot as two separate lines with a gap at lag 0.
    """
    if split:
        filt_neg = f'{-max_lag} <= {lag_key} < 0'
        filt_pos = f'0 < {lag_key} <= {max_lag}'
        g = sns.FacetGrid(dropna=True, **facet_kws, data=recall.reset_index())
        g.map_dataframe(
            lambda data, **kws: sns.lineplot(
                data=data.query(filt_neg), x=lag_key, y='prob', **kws
            )
        )
        g.map_dataframe(
            lambda data, **kws: sns.lineplot(
                data=data.query(filt_pos), x=lag_key, y='prob', **kws
            )
        )
    else:
        data = recall.query(f'{-max_lag} <= {lag_key} <= {max_lag}')
        g = sns.FacetGrid(dropna=False, **facet_kws, data=data.reset_index())
        g.map_dataframe(sns.lineplot, x=lag_key, y='prob')

    g.set_xlabels('Lag')
    g.set_ylabels('CRP')
    g.set(ylim=(0, 1))
    return g


def plot_distance_crp(crp, min_samples=None, **facet_kws):
    """
    Plot response probability by distance bin.

    Parameters
    ----------
    crp : pandas.DataFrame
        Results from `fr.distance_crp`.

    min_samples : int
        Minimum number of samples a bin must have per subject to
        include in the plot.

    **facet_kws
        Additional inputs to pass to `seaborn.relplot`.
    """
    crp = crp.reset_index()
    if min_samples is not None:
        min_n = crp.groupby('center')['possible'].min()
        include = min_n.loc[min_n >= min_samples].index.to_numpy()
        crp = crp.loc[crp['center'].isin(include)]
    g = sns.FacetGrid(dropna=False, **facet_kws, data=crp.reset_index())
    g.map_dataframe(sns.lineplot, x='center', y='prob')
    g.set_xlabels('Distance')
    g.set_ylabels('CRP')
    g.set(ylim=(0, 1))
    return g


def plot_swarm_error(
    data, x=None, y=None, swarm_color=None, swarm_size=5, point_color='k', **facet_kws
):
    """
    Plot points as a swarm plus mean with error bars.

    Parameters
    ----------
    data : pandas.DataFrame
        DataFrame with statistics to plot.

    x : str
        Name of variable to plot on x-axis.

    y : str
        Name of variable to plot on y-axis.

    swarm_color
        Color for swarm plot points. May use any specification
        supported by seaborn.

    swarm_size : float
        Size of swarm plot points.

    point_color
        Color for the point plot (error bars).

    facet_kws
        Additional keywords for the FacetGrid.
    """
    g = sns.FacetGrid(data=data.reset_index(), dropna=False, **facet_kws)
    g.map_dataframe(
        sns.swarmplot, x=x, y=y, color=swarm_color, size=swarm_size, zorder=1
    )
    g.map_dataframe(sns.pointplot, x=x, y=y, color=point_color, join=False, capsize=0.5)
    return g


def plot_raster(
    df,
    hue='input',
    palette=None,
    marker='s',
    intrusion_color=None,
    orientation='horizontal',
    length=6,
    aspect=None,
    legend='auto',
    **facet_kws,
):
    """
    Plot recalls in a raster plot.

    Parameters
    ----------
    df : pandas.DataFrame
        Scored free recall data.

    hue : str or None, optional
        Column to use to set marker color.

    palette : optional
        Palette specification supported by Seaborn.

    marker : str, optional
         Marker code supported by Seaborn.

    intrusion_color : optional
        Color of intrusions.

    orientation : {'horizontal', 'vertical'}, optional
        Whether lists should be stacked horizontally or vertically in
        the plot.

    length : float, optional
        Size of the plot dimension along which list varies.

    aspect : float, optional
        Aspect ratio of plot for lists over items.

    legend : str, optional
        Legend setting. See seaborn.scatterplot for details.

    facet_kws : optional
        Additional key words to pass to seaborn.FacetGrid.
    """
    n_item = int(df['input'].max())
    n_list = int(df['list'].max())
    if palette is None and hue == 'input':
        palette = 'viridis'

    if intrusion_color is None:
        intrusion_color = (0.8, 0.1, 0.3)

    list_lim = (0, n_list + 1)
    item_lim = (0, n_item + 1)
    if orientation == 'horizontal':
        x_var, y_var = 'list', 'output'
        x_lim, y_lim = list_lim, item_lim
        x_label, y_label = 'List', 'Output position'
        def_aspect = n_list / n_item
    else:
        x_var, y_var = 'output', 'list'
        x_lim, y_lim = item_lim, list_lim[::-1]
        x_label, y_label = 'Output position', 'List'
        def_aspect = n_item / n_list

    if aspect is None:
        aspect = def_aspect

    if orientation == 'horizontal':
        height = length / aspect
    else:
        height = length

    g = sns.FacetGrid(
        data=df.reset_index(), dropna=False, aspect=aspect, height=height, **facet_kws
    )
    g.map_dataframe(
        sns.scatterplot,
        x=x_var,
        y=y_var,
        marker=marker,
        hue=hue,
        palette=palette,
        legend=legend,
    )
    g.map_dataframe(
        lambda data, color=None, label=None: sns.scatterplot(
            data=data.query('intrusion'),
            x=x_var,
            y=y_var,
            color=intrusion_color,
            marker=marker,
        )
    )
    g.set_xlabels(x_label)
    g.set_ylabels(y_label)
    g.set(xlim=x_lim, ylim=y_lim)
    return g



"""Test plotting functions."""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import pytest

from psifr import fr


@pytest.fixture()
def data():
    """Create merged free recall data."""
    raw = pd.DataFrame(
        {
            'subject': [
                1, 1, 1, 1, 1, 1,
                1, 1, 1, 1, 1, 1,
            ],
            'list': [
                1, 1, 1, 1, 1, 1,
                2, 2, 2, 2, 2, 2,
            ],
            'trial_type': [
                'study', 'study', 'study', 'recall', 'recall', 'recall',
                'study', 'study', 'study', 'recall', 'recall', 'recall',
            ],
            'position': [
                1, 2, 3, 1, 2, 3,
                1, 2, 3, 1, 2, 3,
            ],
            'item': [
                'absence', 'hollow', 'pupil', 'pupil', 'absence', 'empty',
                'fountain', 'piano', 'pillow', 'pillow', 'fountain', 'pillow',
            ],
            'item_index': [
                0, 1, 2, 1, 2, np.nan,
                3, 4, 5, 5, 3, 5,
            ],
            'task': [
                1, 2, 1, 2, 1, np.nan,
                1, 2, 1, 1, 1, 1,
            ],
        }
    )
    data = fr.merge_free_recall(raw, study_keys=['task'], list_keys=['item_index'])
    return data


@pytest.fixture()
def distances():
    """Create a pairwise distance matrix."""
    mat = np.array(
        [
            [0, 1, 2, 2, 2, 2],
            [1, 0, 1, 2, 2, 2],
            [2, 1, 0, 2, 2, 2],
            [2, 2, 2, 0, 2, 3],
            [2, 2, 2, 2, 0, 2],
            [2, 2, 2, 3, 2, 0],
        ],
    )
    return mat


def test_plot_spc(data):
    """Test plotting a serial position curve."""
    recall = fr.spc(data)
    g = fr.plot_spc(recall)
    plt.close()
    assert isinstance(g, sns.FacetGrid)


def test_plot_lag_crp(data):
    """Test plotting a lag-CRP curve."""
    stat = fr.lag_crp(data)
    g = fr.plot_lag_crp(stat)
    plt.close()
    assert isinstance(g, sns.FacetGrid)


def test_plot_distance_crp(data, distances):
    """Test plotting a distance CRP curve."""
    edges = [0.5, 1.5, 2.5, 3.5]
    stat = fr.distance_crp(data, 'item_index', distances, edges)
    g = fr.plot_distance_crp(stat, min_samples=2)
    plt.close()
    assert isinstance(g, sns.FacetGrid)


def test_plot_swarm_error():
    """Test creating swarm plot with error bars."""
    stat = pd.DataFrame(
        {
            'category': [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2],
            'value': [1, 2, 3, 4, 5, 6, 4, 5, 6, 7, 8, 9],
        }
    )
    g = fr.plot_swarm_error(stat, x='category', y='value')
    plt.close()
    assert isinstance(g, sns.FacetGrid)


def test_plot_raster(data):
    """Test creating a raster plot."""
    g = fr.plot_raster(data)
    plt.close()
    assert isinstance(g, sns.FacetGrid)

    g = fr.plot_raster(data, orientation='vertical')
    plt.close()
    assert isinstance(g, sns.FacetGrid)



"""Test counting transition item pairs."""

import pytest
import numpy as np

from psifr import stats


@pytest.fixture()
def data():
    """Create list data with item numbers."""
    list_data = {
        'n_item': 8,
        'pool_items': [list(range(0, 4)), list(range(4, 8))],
        'recall_items': [[0, 2, 3, 0], [7, 4, 5, 6]],
    }
    return list_data


def test_pair_count(data):
    """Test transition counts by item pairs."""
    actual, possible = stats.count_pairs(
        data['n_item'], data['pool_items'], data['recall_items']
    )
    actual_expected = np.array(
        [
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
        ]
    )
    possib_expected = np.array(
        [
            [0, 1, 1, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 1, 1, 0],
        ]
    )
    np.testing.assert_array_equal(actual, actual_expected)
    np.testing.assert_array_equal(possible, possib_expected)



"""Calculate statistics of list recall."""
import itertools

import numpy
import numpy as np
import pandas as pd
import scipy.stats

from psifr import maskers


def percentile_rank(actual, possible):
    """
    Get percentile rank of a score compared to possible scores.

    Parameters
    ----------
    actual : float
        Score to be ranked. Generally a distance score.

    possible : numpy.ndarray or list
        Possible scores to be compared to.

    Returns
    -------
    rank : float
        Rank scaled to range from 0 (low score) to 1 (high score).

    Examples
    --------
    >>> from psifr import stats
    >>> actual = 3
    >>> possible = [1, 2, 2, 2, 3]
    >>> stats.percentile_rank(actual, possible)
    1.0
    """
    possible_rank = scipy.stats.rankdata(possible)
    actual_rank = possible_rank[actual == np.asarray(possible)]
    possible_count = np.count_nonzero(~np.isnan(possible))
    if possible_count == 1:
        return np.nan
    rank = (actual_rank - 1) / (possible_count - 1)
    return rank[0]


def count_outputs(
    list_length,
    pool_items,
    recall_items,
    pool_label,
    recall_label,
    pool_test=None,
    recall_test=None,
    test=None,
    count_unique=False,
):
    """
    Count actual and possible recalls for each output position.

    Parameters
    ----------
    list_length : int
        Number of items in each list.

    pool_items : list
        List of the serial positions available for recall in each list.
        Must match the serial position codes used in `recall_items`.

    recall_items : list
        List indicating the serial position of each recall in output
        order (NaN for intrusions).

    pool_label : list
        List of the positions to use for calculating lag. Default is to
        use `pool_items`.

    recall_label : list
        List of position labels in recall order. Default is to use
        `recall_items`.

    pool_test : list, optional
         List of some test value for each item in the pool.

    recall_test : list, optional
        List of some test value for each recall attempt by output
        position.

    test : callable
        Callable that evaluates each transition between items n and
        n+1. Must take test values for items n and n+1 and return True
        if a given transition should be included.

    count_unique : bool
        If true, possible recalls with the same label will only be
        counted once.

    Returns
    -------
    actual : numpy.ndarray
        [outputs x inputs] array of actual recall counts.

    possible : numpy.ndarray
        [outputs x inputs] array of possible recall counts.

    Examples
    --------
    >>> from psifr import stats
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> actual, possible = stats.count_outputs(
    ...     4, pool_items, recall_items, pool_items, recall_items
    ... )
    >>> actual
    array([[0, 0, 0, 1],
           [0, 1, 0, 0],
           [0, 0, 1, 0],
           [1, 0, 0, 0]])
    >>> possible
    array([[1, 1, 1, 1],
           [1, 1, 1, 0],
           [1, 0, 1, 0],
           [1, 0, 0, 0]])
    """
    if pool_label is None:
        pool_label = pool_items

    if recall_label is None:
        recall_label = recall_items

    count_actual = np.zeros((list_length, list_length), dtype=int)
    count_possible = np.zeros((list_length, list_length), dtype=int)
    for i, recall_items_list in enumerate(recall_items):
        # set up masker to filter outputs
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.outputs_masker(
            pool_items[i],
            recall_items_list,
            pool_label[i],
            recall_label[i],
            pool_test_list,
            recall_test_list,
            test,
        )

        for curr, poss, op in masker:
            curr = int(curr)
            poss = poss.astype(int)

            # for this step, calculate actual input position and
            # possible input positions
            count_actual[op - 1, curr - 1] += 1
            if count_unique:
                for j in poss:
                    count_possible[op - 1, j - 1] += 1
            else:
                count_possible[op - 1, poss - 1] += 1
    return count_actual, count_possible


def count_lags(
    list_length,
    pool_items,
    recall_items,
    pool_label=None,
    recall_label=None,
    pool_test=None,
    recall_test=None,
    test=None,
    count_unique=False,
):
    """
    Count actual and possible serial position lags.

    Parameters
    ----------
    list_length : int
        Number of items in each list.

    pool_items : list
        List of the serial positions available for recall in each list.
        Must match the serial position codes used in `recall_items`.

    recall_items : list
        List indicating the serial position of each recall in output
        order (NaN for intrusions).

    pool_label : list, optional
        List of the positions to use for calculating lag. Default is to
        use `pool_items`.

    recall_label : list, optional
        List of position labels in recall order. Default is to use
        `recall_items`.

    pool_test : list, optional
         List of some test value for each item in the pool.

    recall_test : list, optional
        List of some test value for each recall attempt by output
        position.

    test : callable
        Callable that evaluates each transition between items n and
        n+1. Must take test values for items n and n+1 and return True
        if a given transition should be included.

    count_unique : bool, optional
        If true, only unique values will be counted toward the possible
        transitions. If multiple items are avilable for recall for a
        given transition and a given bin, that bin will only be
        incremented once. If false, all possible transitions will add
        to the count.

    Returns
    -------
    actual : pandas.Series
        Count of actual lags that occurred in the recall sequence.

    possible : pandas.Series
        Count of possible lags.

    See Also
    --------
    rank_lags : Rank of serial position lags.

    Examples
    --------
    >>> from psifr import stats
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> actual, possible = stats.count_lags(4, pool_items, recall_items)
    >>> actual
    lag
    -3    0
    -2    2
    -1    0
     0    0
     1    1
     2    0
     3    0
    dtype: int64
    >>> possible
    lag
    -3    1
    -2    2
    -1    2
     0    0
     1    1
     2    0
     3    0
    dtype: int64
    """
    if pool_label is None:
        pool_label = pool_items

    if recall_label is None:
        recall_label = recall_items

    list_actual = []
    list_possible = []
    for i, recall_items_list in enumerate(recall_items):
        # set up masker to filter transitions
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items_list,
            pool_label[i],
            recall_label[i],
            pool_test_list,
            recall_test_list,
            test,
        )

        for output, prev, curr, poss in masker:
            # for this step, calculate actual lag and all possible lags
            list_actual.append(curr - prev)
            tran_poss = poss - prev
            if count_unique:
                tran_poss = np.unique(tran_poss)
            list_possible.extend(tran_poss)

    # count the actual and possible transitions for each lag
    max_lag = list_length - 1
    lags = np.arange(-max_lag, max_lag + 2)
    index = pd.Index(lags[:-1], name='lag')
    actual = pd.Series(np.histogram(list_actual, lags)[0], index=index)
    possible = pd.Series(np.histogram(list_possible, lags)[0], index=index)
    return actual, possible


def count_lags_compound(
    list_length,
    pool_items,
    recall_items,
    pool_label=None,
    recall_label=None,
    pool_test=None,
    recall_test=None,
    test=None,
    count_unique=False,
):
    """
    Count lags conditional on the lag of the previous transition.

    Parameters
    ----------
    list_length : int
        Number of items in each list.

    pool_items : list
        List of the serial positions available for recall in each list.
        Must match the serial position codes used in `recall_items`.

    recall_items : list
        List indicating the serial position of each recall in output
        order (NaN for intrusions).

    pool_label : list, optional
        List of the positions to use for calculating lag. Default is to
        use `pool_items`.

    recall_label : list, optional
        List of position labels in recall order. Default is to use
        `recall_items`.

    pool_test : list, optional
         List of some test value for each item in the pool.

    recall_test : list, optional
        List of some test value for each recall attempt by output
        position.

    test : callable
        Callable that evaluates each transition between items n and
        n+1. Must take test values for items n and n+1 and return True
        if a given transition should be included.

    count_unique : bool, optional
        If true, only unique values will be counted toward the possible
        transitions. If multiple items are avilable for recall for a
        given transition and a given bin, that bin will only be
        incremented once. If false, all possible transitions will add
        to the count.

    Returns
    -------
    actual : pandas.Series
        Count of actual lags that occurred in the recall sequence.

    possible : pandas.Series
        Count of possible lags.

    See Also
    --------
    count_lags : Count of individual transitions.

    Examples
    --------
    >>> from psifr import stats
    >>> pool_items = [[1, 2, 3]]
    >>> recall_items = [[3, 1, 2]]
    >>> actual, possible = stats.count_lags_compound(3, pool_items, recall_items)
    >>> (actual == possible).all()
    True
    >>> actual
    previous  current
    -2        -2         0
              -1         0
               0         0
               1         1
               2         0
    -1        -2         0
              -1         0
               0         0
               1         0
               2         0
     0        -2         0
              -1         0
               0         0
               1         0
               2         0
     1        -2         0
              -1         0
               0         0
               1         0
               2         0
     2        -2         0
              -1         0
               0         0
               1         0
               2         0
    dtype: int64
    """
    if pool_label is None:
        pool_label = pool_items

    if recall_label is None:
        recall_label = recall_items

    prev_actual = []
    prev_possible = []
    curr_actual = []
    curr_possible = []
    for i, recall_items_list in enumerate(recall_items):
        # set up masker to filter pairs of transitions
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.sequences_masker(
            2,
            pool_items[i],
            recall_items_list,
            pool_label[i],
            recall_label[i],
            pool_test_list,
            recall_test_list,
            test,
        )

        for output, prev, curr, poss in masker:
            # calculate lag of prior transition
            prev_lag = curr[-2] - prev[-2]

            # actual and possible lags of current transition
            curr_lag = curr[-1] - prev[-1]
            poss_lag = poss[-1] - prev[-1]

            # add lags to count
            prev_actual.append(prev_lag)
            curr_actual.append(curr_lag)
            if count_unique:
                poss_lag = np.unique(poss_lag)
            prev_possible.extend([prev_lag] * len(poss_lag))
            curr_possible.extend(poss_lag)

    # define possible combinations of previous and current lags
    max_lag = list_length - 1
    bins = np.arange(-max_lag, max_lag + 2)
    lags = bins[:-1]
    compound_lags = list(itertools.product(lags, lags))
    index = pd.MultiIndex.from_tuples(compound_lags, names=['previous', 'current'])

    # count the actual and possible transitions for each (lag, lag)
    # combination
    count_actual = np.histogram2d(prev_actual, curr_actual, bins)[0].astype(int)
    actual = pd.Series(count_actual.ravel(), index=index)
    count_possible = np.histogram2d(prev_possible, curr_possible, bins)[0].astype(int)
    possible = pd.Series(count_possible.ravel(), index=index)
    return actual, possible


def rank_lags(
    pool_items,
    recall_items,
    pool_label=None,
    recall_label=None,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Calculate rank of absolute lag for free recall lists.

    Parameters
    ----------
    pool_items : list
        List of the serial positions available for recall in each list.
        Must match the serial position codes used in `recall_items`.

    recall_items : list
        List indicating the serial position of each recall in output
        order (NaN for intrusions).

    pool_label : list, optional
        List of the positions to use for calculating lag. Default is to
        use `pool_items`.

    recall_label : list, optional
        List of position labels in recall order. Default is to use
        `recall_items`.

    pool_test : list, optional
         List of some test value for each item in the pool.

    recall_test : list, optional
        List of some test value for each recall attempt by output
        position.

    test : callable
        Callable that evaluates each transition between items n and
        n+1. Must take test values for items n and n+1 and return True
        if a given transition should be included.

    Returns
    -------
    rank : list
        Absolute lag percentile rank for each included transition. The
        rank is 0 if the lag was the most distant of the available
        transitions, and 1 if the lag was the closest. Ties are
        assigned to the average percentile rank.

    See Also
    --------
    count_lags : Count actual and possible serial position lags.

    Examples
    --------
    >>> from psifr import stats
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> stats.rank_lags(pool_items, recall_items)
    [0.5, 0.5, nan]
    """
    if pool_label is None:
        pool_label = pool_items

    if recall_label is None:
        recall_label = recall_items

    rank = []
    for i, recall_items_list in enumerate(recall_items):
        # set up masker to filter transitions
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items_list,
            pool_label[i],
            recall_label[i],
            pool_test_list,
            recall_test_list,
            test,
        )

        for output, prev, curr, poss in masker:
            actual = np.abs(curr - prev)
            possible = np.abs(poss - prev)
            rank.append(1 - percentile_rank(actual, possible))
    return rank


def count_distance(
    distances,
    edges,
    pool_items,
    recall_items,
    pool_index,
    recall_index,
    pool_test=None,
    recall_test=None,
    test=None,
    count_unique=False,
):
    """
    Count transitions within distance bins.

    Parameters
    ----------
    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    edges : array-like
        Edges of bins to apply to distances.

    pool_items : list of list
        Unique item codes for each item in the pool available for recall.

    recall_items : list of list
        Unique item codes of recalled items.

    pool_index : list of list
        Index of each item in the distances matrix.

    recall_index : list of list
        Index of each recalled item.

    pool_test : list of list, optional
        Test value for each item in the pool.

    recall_test : list of list, optional
        Test value for each recalled item.

    test : callable
        Called as test(prev, curr) or test(prev, poss) to screen
        actual and possible transitions, respectively.

    count_unique : bool, optional
        If true, only unique values will be counted toward the possible
        transitions. If multiple items are avilable for recall for a
        given transition and a given bin, that bin will only be
        incremented once. If false, all possible transitions will add
        to the count.

    Returns
    -------
    actual : pandas.Series
        Count of actual transitions made for each bin.

    possible : pandas.Series
        Count of possible transitions for each bin.

    See Also
    --------
    rank_distance : Calculate percentile rank of transition distances.

    Examples
    --------
    >>> import numpy as np
    >>> from psifr import stats
    >>> distances = np.array([[0, 1, 2, 2], [1, 0, 2, 2], [2, 2, 0, 3], [2, 2, 3, 0]])
    >>> edges = np.array([0.5, 1.5, 2.5, 3.5])
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> pool_index = [[0, 1, 2, 3]]
    >>> recall_index = [[3, 1, 2, 0]]
    >>> actual, possible = stats.count_distance(
    ...     distances, edges, pool_items, recall_items, pool_index, recall_index
    ... )
    >>> actual
    (0.5, 1.5]    0
    (1.5, 2.5]    3
    (2.5, 3.5]    0
    Name: count, dtype: int64
    >>> possible
    (0.5, 1.5]    1
    (1.5, 2.5]    4
    (2.5, 3.5]    1
    Name: count, dtype: int64
    """
    list_actual = []
    list_possible = []
    centers = edges[:-1] + np.diff(edges) / 2
    for i in range(len(recall_items)):
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items[i],
            pool_index[i],
            recall_index[i],
            pool_test_list,
            recall_test_list,
            test,
        )
        for output, prev, curr, poss in masker:
            prev = int(prev)
            curr = int(curr)
            poss = poss.astype(int)

            list_actual.append(distances[prev, curr])
            tran_poss = distances[prev, poss]
            if count_unique:
                # get count of each possible bin
                bin_count_poss = np.histogram(tran_poss, edges)[0]

                # for each bin that was possible, add the center as a
                # possible transition
                tran_poss = centers[np.nonzero(bin_count_poss)[0]]
            list_possible.extend(tran_poss)

    actual = pd.cut(list_actual, edges).value_counts()
    possible = pd.cut(list_possible, edges).value_counts()
    return actual, possible


def rank_distance(
    distances,
    pool_items,
    recall_items,
    pool_index,
    recall_index,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Calculate percentile rank of transition distances.

    Parameters
    ----------
    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    pool_items : list of list
        Unique item codes for each item in the pool available for recall.

    recall_items : list of list
        Unique item codes of recalled items.

    pool_index : list of list
        Index of each item in the distances matrix.

    recall_index : list of list
        Index of each recalled item.

    pool_test : list of list, optional
        Test value for each item in the pool.

    recall_test : list of list, optional
        Test value for each recalled item.

    test : callable
        Called as test(prev, curr) or test(prev, poss) to screen
        actual and possible transitions, respectively.

    Returns
    -------
    rank : list
        Distance percentile rank for each included transition. The
        rank is 0 if the distance was the largest of the available
        transitions, and 1 if the distance was the smallest. Ties are
        assigned to the average percentile rank.

    See Also
    --------
    count_distance : Count transitions within distance bins.

    Examples
    --------
    >>> import numpy as np
    >>> from psifr import stats
    >>> distances = np.array([[0, 1, 2, 2], [1, 0, 2, 2], [2, 2, 0, 3], [2, 2, 3, 0]])
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> pool_index = [[0, 1, 2, 3]]
    >>> recall_index = [[3, 1, 2, 0]]
    >>> stats.rank_distance(
    ...     distances, pool_items, recall_items, pool_index, recall_index
    ... )
    [0.75, 0.0, nan]
    """
    rank = []
    for i in range(len(recall_items)):
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items[i],
            pool_index[i],
            recall_index[i],
            pool_test_list,
            recall_test_list,
            test,
        )
        for output, prev, curr, poss in masker:
            prev = int(prev)
            curr = int(curr)
            poss = poss.astype(int)
            actual = distances[prev, curr]
            possible = distances[prev, poss]
            rank.append(1 - percentile_rank(actual, possible))
    return rank


def rank_distance_shifted(
    distances,
    max_shift,
    pool_items,
    recall_items,
    pool_index,
    recall_index,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Calculate percentile rank of shifted distances.

    Parameters
    ----------
    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    max_shift : int
        Maximum number of items back for which to rank distances.

    pool_items : list of list
        Unique item codes for each item in the pool available for recall.

    recall_items : list of list
        Unique item codes of recalled items.

    pool_index : list of list
        Index of each item in the distances matrix.

    recall_index : list of list
        Index of each recalled item.

    pool_test : list of list, optional
        Test value for each item in the pool.

    recall_test : list of list, optional
        Test value for each recalled item.

    test : callable
        Called as test(prev, curr) or test(prev, poss) to screen
        actual and possible transitions, respectively.

    Returns
    -------
    rank : numpy.ndarray
        [transitions x max_shift] array with distance percentile ranks.
        The rank is 0 if the distance was the largest of the available
        transitions, and 1 if the distance was the smallest. Ties are
        assigned to the average percentile rank.

    See Also
    --------
    rank_distance - Percentile rank of transition distances relative
        to the immediately preceding item only.

    Examples
    --------
    >>> import numpy as np
    >>> from psifr import stats
    >>> distances = np.array(
    ...     [
    ...         [0, 1, 2, 2, 2],
    ...         [1, 0, 2, 2, 2],
    ...         [2, 2, 0, 3, 3],
    ...         [2, 2, 3, 0, 2],
    ...         [2, 2, 3, 2, 0],
    ...     ]
    ... )
    >>> pool_items = [[1, 2, 3, 4, 5]]
    >>> recall_items = [[4, 2, 3, 1]]
    >>> pool_index = [[0, 1, 2, 3, 4]]
    >>> recall_index = [[3, 1, 2, 0]]
    >>> stats.rank_distance_shifted(
    ...     distances, 2, pool_items, recall_items, pool_index, recall_index
    ... )
    array([[0.  , 0.25],
           [1.  , 1.  ]])
    """
    rank = []
    for i in range(len(recall_items)):
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.sequences_masker(
            max_shift,
            pool_items[i],
            recall_items[i],
            pool_index[i],
            recall_index[i],
            pool_test_list,
            recall_test_list,
            test,
        )
        for s_output, s_prev, s_curr, s_poss in masker:
            curr = int(s_curr[-1])
            poss = s_poss[-1].astype(int)
            rank_shift = []
            for j in range(max_shift, 0, -1):
                # rank previous to current distance for this shift
                prev = int(s_prev[-j])
                actual = distances[prev, curr]
                possible = distances[prev, poss]
                rank_shift.append(1 - percentile_rank(actual, possible))
            rank.append(rank_shift)
    rank = np.array(rank)
    return rank


def rank_distance_window(
    distances,
    list_length,
    window_lags,
    pool_items,
    recall_items,
    pool_index,
    recall_index,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Calculate percentile rank of distances relative to items in a window.

    Parameters
    ----------
    distances : numpy.array
        Items x items matrix of pairwise distances or similarities.

    list_length : int
        Number of items in each list.

    window_lags : array_like
        Serial position lags to include in the window.

    pool_items : list of list
        Unique item codes for each item in the pool available for recall.

    recall_items : list of list
        Unique item codes of recalled items.

    pool_index : list of list
        Index of each item in the distances matrix.

    recall_index : list of list
        Index of each recalled item.

    pool_test : list of list, optional
        Test value for each item in the pool.

    recall_test : list of list, optional
        Test value for each recalled item.

    test : callable
        Called as test(prev, curr) or test(prev, poss) to screen
        actual and possible transitions, respectively.

    Returns
    -------
    rank : numpy.ndarray
        [transitions x window lags] array with distance percentile ranks.
        The rank is 0 if the distance was the largest of the available
        transitions, and 1 if the distance was the smallest. Ties are
        assigned to the average percentile rank.
    """
    rank = []
    for i in range(len(recall_items)):
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.windows_masker(
            list_length,
            window_lags,
            pool_items[i],
            recall_items[i],
            pool_index[i],
            recall_index[i],
            pool_test_list,
            recall_test_list,
            test,
        )
        for output, w_prev, curr, poss in masker:
            curr = int(curr)
            rank_lag = []
            for prev in w_prev:
                prev = int(prev)
                poss = poss.astype(int)
                actual = distances[prev, curr]
                possible = distances[prev, poss]
                rank_lag.append(1 - percentile_rank(actual, possible))
            rank.append(rank_lag)
    rank = np.array(rank)
    return rank


def count_category(
    pool_items,
    recall_items,
    pool_category,
    recall_category,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Count within-category transitions.

    Parameters
    ----------
    pool_items : list
        List of the serial positions available for recall in each list.
        Must match the serial position codes used in `recall_items`.

    recall_items : list
        List indicating the serial position of each recall in output
        order (NaN for intrusions).

    pool_category : list
        List of the category of each item in the pool for each list.

    recall_category : list
        List of item category in recall order.

    pool_test : list, optional
         List of some test value for each item in the pool.

    recall_test : list, optional
        List of some test value for each recall attempt by output
        position.

    test : callable
        Callable that evaluates each transition between items n and
        n+1. Must take test values for items n and n+1 and return True
        if a given transition should be included.

    Returns
    -------
    actual : int
        Count of actual within-category transitions.

    possible : int
        Count of possible within-category transitions.

    Examples
    --------
    >>> from psifr import stats
    >>> pool_items = [[1, 2, 3, 4]]
    >>> recall_items = [[4, 3, 1, 2]]
    >>> pool_category = [[1, 1, 2, 2]]
    >>> recall_category = [[2, 2, 1, 1]]
    >>> stats.count_category(
    ...     pool_items, recall_items, pool_category, recall_category
    ... )
    (2, 2)
    """
    actual = 0
    possible = 0
    for i in range(len(recall_items)):
        # set up masker to filter transitions
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items[i],
            pool_category[i],
            recall_category[i],
            pool_test_list,
            recall_test_list,
            test,
        )

        for output, prev, curr, poss in masker:
            if prev == curr:
                actual += 1
            if np.any(prev == poss):
                possible += 1
    return actual, possible


def count_pairs(
    n_item, pool_items, recall_items, pool_test=None, recall_test=None, test=None
):
    """Count transitions between pairs of specific items."""
    actual = np.zeros((n_item, n_item), dtype=int)
    possible = np.zeros((n_item, n_item), dtype=int)
    for i, recall_items_list in enumerate(recall_items):
        # set up masker to filter transitions
        pool_test_list = None if pool_test is None else pool_test[i]
        recall_test_list = None if recall_test is None else recall_test[i]
        masker = maskers.transitions_masker(
            pool_items[i],
            recall_items_list,
            pool_items[i],
            recall_items_list,
            pool_test_list,
            recall_test_list,
            test,
        )

        for output, prev, curr, poss in masker:
            actual[prev, curr] += 1
            possible[prev, poss] += 1
    return actual, possible


def lbc(study_category, recall_category):
    """Calculate list-based clustering (LBC) for a set of lists."""
    lbc_scores = np.zeros(len(study_category))
    for i, (study, recall) in enumerate(zip(study_category, recall_category)):
        study = np.asarray(study)
        recall = np.asarray(recall)
        if np.any(pd.isna(study)):
            raise ValueError('Study category contains N/A values.')
        if np.any(pd.isna(recall)):
            raise ValueError('Recall category contains N/A values.')

        # number of correct recalls
        r = len(recall)

        # number of items per category
        m = len(study) / len(np.unique(study))

        # list length
        nl = len(study)

        # observed and expected clustering
        observed = np.count_nonzero(recall[:-1] == recall[1:])
        expected = ((r - 1) * (m - 1)) / (nl - 1)
        lbc_scores[i] = observed - expected
    return lbc_scores


def arc(recall_category):
    """Calculate adjusted ratio of clustering for a set of lists."""
    arc_scores = np.zeros(len(recall_category))
    for i, recall in enumerate(recall_category):
        recall = np.asarray(recall)
        if np.any(pd.isna(recall)):
            raise ValueError('Recall category contains N/A values.')

        # number of categories and correct recalls from each category
        categories = np.unique(recall)
        n = np.array([np.count_nonzero(recall == c) for c in categories])
        c = len(categories)

        # number of correct recalls
        r = len(recall)

        # observed, expected, and maximum clustering
        expected = np.sum((n * (n - 1)) / r)
        observed = np.count_nonzero(recall[:-1] == recall[1:])
        maximum = r - c
        if maximum == expected:
            # when maximum is the same as expected, ARC is undefined
            arc_scores[i] = np.nan
            continue
        arc_scores[i] = (observed - expected) / (maximum - expected)
    return arc_scores



"""Test counting recalls by output."""

import numpy as np
import pytest

from psifr import stats


@pytest.fixture()
def list_data():
    """Create lists with position and category."""
    data = {
        'list_length': 4,
        'pool_items': [[1, 2, 3, 4], [1, 2, 3, 4]],
        'recall_items': [[4, 3], [4, 1, 2, 3]],
        'pool_test': [[1, 1, 2, 2], [1, 1, 2, 2]],
        'recall_test': [[2, 2], [2, 1, 1, 2]],
    }
    return data


def test_count_output(list_data):
    """Test actual and possible serial positions by output position."""
    actual, possible = stats.count_outputs(
        list_data['list_length'],
        list_data['pool_items'],
        list_data['recall_items'],
        list_data['pool_items'],
        list_data['recall_items'],
    )
    expected_actual = np.array(
        [
            [0, 0, 0, 2],
            [1, 0, 1, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
        ]
    )
    expected_possible = np.array(
        [
            [2, 2, 2, 2],
            [2, 2, 2, 0],
            [0, 1, 1, 0],
            [0, 0, 1, 0],
        ]
    )
    np.testing.assert_array_equal(actual, expected_actual)
    np.testing.assert_array_equal(possible, expected_possible)



"""Utilities to iterate over masked lists of recalls."""

import numpy as np
import pandas as pd


def outputs_masker(
    pool_items,
    recall_items,
    pool_output,
    recall_output,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Iterate over valid outputs.

    Parameters
    ----------
    pool_items : list
        Items available for recall. Order does not matter. May contain
        repeated values. Item identifiers must be unique within pool.

    recall_items : list
        Recalled items in output position order.

    pool_output : list
        Output values for pool items. Must be the same order as pool.

    recall_output : list
        Output values in output position order.

    pool_test : list, optional
        Test values for items available for recall. Must be the same
        order as pool.

    recall_test : list, optional
        Test values for items in output position order.

    test : callable, optional
        Used to test whether output recalls and possible recalls should
        be included, based on their test values.

    Yields
    ------
    curr : object
        Output value for the item at this valid output position.

    poss : numpy.array
        Output values for all possible items that could be recalled at
        this output position.

    output : int
        Current output position.

    Examples
    --------
    >>> from psifr import maskers
    >>> pool_items = [1, 2, 3, 4]
    >>> recall_items = [4, 2, 3, 1]
    >>> masker = maskers.outputs_masker(
    ...     pool_items, recall_items, pool_items, recall_items
    ... )
    >>> for curr, poss, output in masker:
    ...     print(curr, poss, output)
    4 [1 2 3 4] 1
    2 [1 2 3] 2
    3 [1 3] 3
    1 [1] 4
    """
    pool_items = pool_items.copy()
    pool_output = pool_output.copy()
    if test is not None:
        pool_test = pool_test.copy()

    n = 0
    output = 0
    while n < len(recall_items):
        # test if the current item is in the pool
        if recall_items[n] not in pool_items:
            n += 1
            continue
        output += 1

        curr = recall_output[n]
        poss = np.array(pool_output)

        # remove the item from the pool
        ind = pool_items.index(recall_items[n])
        del pool_items[ind]
        del pool_output[ind]
        if test is not None:
            del pool_test[ind]

        if test is not None:
            # test if this recall is included
            if not test(recall_test[n]):
                n += 1
                continue

            # get included possible items
            poss = poss[test(np.array(pool_test))]
        n += 1
        yield curr, poss, output


def transitions_masker(
    pool_items,
    recall_items,
    pool_output,
    recall_output,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Iterate over transitions with masking.

    Transitions are between a "previous" item and a "current" item.
    Non-included transitions will be skipped. A transition is yielded
    only if it matches the following conditions:

    (1) Each item involved in the transition is in the pool. Items are
    removed from the pool after they appear as the previous item.

    (2) Optionally, an additional check is run based on test values
    associated with the items in the transition. For example, this
    could be used to only include transitions where the category of
    the previous and current items is the same.

    The masker will yield "output" values, which may be distinct from
    the item identifiers used to determine item repeats.

    Parameters
    ----------
    pool_items : list
        Items available for recall. Order does not matter. May contain
        repeated values. Item identifiers must be unique within pool.

    recall_items : list
        Recalled items in output position order.

    pool_output : list
        Output values for pool items. Must be the same order as pool.

    recall_output : list
        Output values in output position order.

    pool_test : list, optional
        Test values for items available for recall. Must be the same
        order as pool.

    recall_test : list, optional
        Test values for items in output position order.

    test : callable, optional
        Used to test whether individual transitions should be included,
        based on test values.

            test(prev, curr) - test for included transition

            test(prev, poss) - test for included possible transition

    Yields
    ------
    output : int
        Output position of this transition. The first transition is 1.

    prev : object
        Output value for the "from" item on this transition.

    curr : object
        Output value for the "to" item.

    poss : numpy.array
        Output values for all possible valid "to" items.

    Examples
    --------
    >>> from psifr import maskers
    >>> pool = [1, 2, 3, 4, 5, 6]
    >>> recs = [6, 2, 3, 6, 1, 4]
    >>> masker = maskers.transitions_masker(
    ...     pool_items=pool, recall_items=recs, pool_output=pool, recall_output=recs
    ... )
    >>> for output, prev, curr, poss in masker:
    ...     print(output, prev, curr, poss)
    1 6 2 [1 2 3 4 5]
    2 2 3 [1 3 4 5]
    5 1 4 [4 5]
    """
    pool_items = pool_items.copy()
    pool_output = pool_output.copy()
    if test is not None:
        pool_test = pool_test.copy()

    for n in range(len(recall_items) - 1):
        # test if the previous item is in the pool
        if pd.isnull(recall_items[n]) or (recall_items[n] not in pool_items):
            continue

        # remove the item from the pool
        ind = pool_items.index(recall_items[n])
        del pool_items[ind]
        del pool_output[ind]
        if test is not None:
            del pool_test[ind]

        # test if the current item is in the pool
        if pd.isnull(recall_items[n + 1]) or (recall_items[n + 1] not in pool_items):
            continue

        prev = recall_output[n]
        curr = recall_output[n + 1]
        poss = np.array(pool_output)
        if test is not None:
            # test if this transition is included
            if not test(recall_test[n], recall_test[n + 1]):
                continue

            # get included possible items
            ind = test(recall_test[n], np.array(pool_test))
            if not isinstance(ind, np.ndarray):
                ind = np.repeat(ind, poss.shape)
            poss = poss[ind]
        yield n + 1, prev, curr, poss


def sequences_masker(
    n_transitions,
    pool_items,
    recall_items,
    pool_output,
    recall_output,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Yield sequences of adjacent included transitions.

    Parameters
    ----------
    n_transitions : int
        Number of transitions to include in yielded sequences.

    pool_items : list
        Items available for recall. Order does not matter. May contain
        repeated values. Item identifiers must be unique within pool.

    recall_items : list
        Recalled items in output position order.

    pool_output : list
        Output values for pool items. Must be the same order as pool.

    recall_output : list
        Output values in output position order.

    pool_test : list, optional
        Test values for items available for recall. Must be the same
        order as pool.

    recall_test : list, optional
        Test values for items in output position order.

    test : callable, optional
        Used to test whether individual transitions should be included,
        based on test values.

            test(prev, curr) - test for included transition

            test(prev, poss) - test for included possible transition

    Yields
    ------
    output : int
        Output positions of included transitions. The first transition
        is 1.

    prev : list
        Output values for the "from" item in included transitions.

    curr : list
        Output values for the "to" item in included transitions.

    poss : list of numpy.ndarray
        Output values for all possible valid "to" items in included
        transitions.

    See Also
    --------
    transitions_masker : Yield included transitions.

    Examples
    --------
    >>> from psifr import maskers
    >>> pool = [1, 2, 3, 4, 5, 6]
    >>> recs = [6, 2, 3, 6, 1, 4, 5]
    >>> masker = maskers.sequences_masker(
    ...     2, pool_items=pool, recall_items=recs, pool_output=pool, recall_output=recs
    ... )
    >>> for output, prev, curr, poss in masker:
    ...     print(output, prev, curr, poss)
    [1, 2] [6, 2] [2, 3] [array([1, 2, 3, 4, 5]), array([1, 3, 4, 5])]
    [5, 6] [1, 4] [4, 5] [array([4, 5]), array([5])]

    >>> pool = [1, 2, 3, 4]
    >>> recs = [4, 3, 1, 2]
    >>> masker = maskers.sequences_masker(
    ...     3, pool_items=pool, recall_items=recs, pool_output=pool, recall_output=recs
    ... )
    >>> for output, prev, curr, poss in masker:
    ...     print(output, prev, curr, poss)
    [1, 2, 3] [4, 3, 1] [3, 1, 2] [array([1, 2, 3]), array([1, 2]), array([2])]
    """
    masker = transitions_masker(
        pool_items,
        recall_items,
        pool_output,
        recall_output,
        pool_test=pool_test,
        recall_test=recall_test,
        test=test,
    )
    s_output = []
    s_prev = []
    s_curr = []
    s_poss = []
    prev_output = 0
    sequence_len = 0
    for output, prev, curr, poss in masker:
        if (output - prev_output) == 1:
            s_output.append(output)
            s_prev.append(prev)
            s_curr.append(curr)
            s_poss.append(poss)
            sequence_len += 1
        else:
            # a break in the chain
            s_output = [output]
            s_prev = [prev]
            s_curr = [curr]
            s_poss = [poss]
            sequence_len = 1
        prev_output = output

        if sequence_len >= n_transitions:
            ind = slice(sequence_len - n_transitions, None)
            yield s_output[ind], s_prev[ind], s_curr[ind], s_poss[ind]


def windows_masker(
    list_length,
    window_lags,
    pool_items,
    recall_items,
    pool_output,
    recall_output,
    pool_test=None,
    recall_test=None,
    test=None,
):
    """
    Yield windows around previous items in the input list.

    Parameters
    ----------
    list_length : int
        Number of items in each list.

    window_lags : array_like
        Serial position lags to include in the window.

    pool_items : list
        Input position of items available for recall.

    recall_items : list
        Input position of recalled items, in output position order.

    pool_output : list
        Output values for pool items. Must be the same order as pool.

    recall_output : list
        Output values in output position order.

    pool_test : list, optional
        Test values for items available for recall. Must be the same
        order as pool.

    recall_test : list, optional
        Test values for items in output position order.

    test : callable, optional
        Used to test whether individual transitions should be included,
        based on test values.

            test(prev, curr) - test for included transition

            test(prev, poss) - test for included possible transition

    Yields
    ------
    output : int
        Output positions of included transitions. The first transition
        is 1.

    prev : list
        Output values for the "from" item in included transitions.

    curr : list
        Output values for the "to" item in included transitions.

    poss : list of numpy.ndarray
        Output values for all possible valid "to" items in included
        transitions.
    """
    # pool items include all items in presentation order; poss items
    # include only items that have not been recalled yet
    poss_items = pool_items.copy()
    poss_output = pool_output.copy()
    pool_output = np.asarray(pool_output)
    if test is not None:
        poss_test = pool_test.copy()
        pool_test = np.asarray(pool_test)
    window_lags = np.asarray(window_lags)

    for n in range(len(recall_items) - 1):
        # test if the previous item is in the pool
        if pd.isnull(recall_items[n]) or (recall_items[n] not in poss_items):
            continue

        # remove the item from the pool
        ind = poss_items.index(recall_items[n])
        del poss_items[ind]
        del poss_output[ind]
        if test is not None:
            del poss_test[ind]

        # test if the current item is in the pool
        if pd.isnull(recall_items[n + 1]) or (recall_items[n + 1] not in poss_items):
            continue

        # get windowed items in the input list
        prev = int(recall_items[n]) + window_lags

        # exclude if any windowed items do not exist or fail test
        if np.any(prev < 1) or np.any(prev > list_length):
            continue

        # exclude current/possible items in the window
        curr = int(recall_items[n + 1])
        if curr in prev:
            continue
        poss = np.asarray(poss_items, int)
        include_poss = ~np.isin(poss, prev)
        poss = poss[include_poss]

        if test is not None:
            # test if this transition is included
            if np.any(~test(pool_test[prev - 1], recall_test[n + 1])):
                continue

            # get included possible items
            include = test(pool_test[prev - 1][:, np.newaxis], pool_test[poss - 1])
            poss = poss[np.all(include, axis=0)]
        yield n + 1, pool_output[prev - 1], pool_output[curr - 1], pool_output[poss - 1]



"""Classes for defining recall measures."""

import abc

import numpy as np
import pandas as pd

from psifr import stats
from psifr import fr


class TransitionMeasure(object):
    """
    Measure of free recall dataset with multiple subjects.

    Parameters
    ----------
    items_key : str
        Data column with item identifiers.

    label_key : str
        Data column with trial labels to use for the measure.

    item_query : str
        Query string to indicate trials to include in the measure.

    test_key : str
        Data column with labels to use when testing for trial inclusion.

    test : callable
        Test of trial inclusion. Takes the previous and current test
        values and return True if the transition should be included.

    Attributes
    ----------
    keys : dict of {str: str}
        List of columns to use for the measure.

    item_query : str
        Query string to indicate trials to include in the measure.

    test : callable
        Test of trial inclusion.
    """

    def __init__(self, items_key, label_key, item_query=None, test_key=None, test=None):

        self.keys = {'items': items_key, 'label': label_key, 'test': test_key}
        self.item_query = item_query
        self.test = test

    def split_lists(self, data, phase, item_query=None):
        """
        Get relevant fields and split by list.

        Parameters
        ----------
        data : pandas.DataFrame
            Raw free recall data.

        phase : str
            Phase to split ('study' or 'recall').

        item_query : str, optional
            Query string to determine included trials.
        """
        names = list(self.keys.keys())
        keys = list(self.keys.values())
        for key in keys:
            if (key is not None) and (key not in data.columns):
                raise ValueError(f'Required column {key} is missing.')
        split = fr.split_lists(data, phase, keys, names, item_query, as_list=True)
        return split

    @abc.abstractmethod
    def analyze_subject(self, subject, pool_lists, recall_lists):
        """
        Analyze a single subject.

        Parameters
        ----------
        subject : int or str
            Identifier of the subject to analyze.

        pool_lists : dict of lists of numpy.ndarray
            Information about the item pool for each list, with keys
            for items, label, and test arrays.

        recall_lists : dict of lists of numpy.ndarray
            Information about the recall sequence for each list, with
            keys for items, label, and test arrays.

        Returns
        -------
        pandas.DataFrame
            Results of the analysis for one subject. Should include a
            'subject' column in the index.
        """
        pass

    def analyze(self, data):
        """
        Analyze a free recall dataset with multiple subjects.

        Parameters
        ----------
        data : pandas.DataFrame
            Raw (not merged) free recall data.

        Returns
        -------
        stat : pandas.DataFrame
            Statistics calculated for each subject.
        """
        subj_results = []
        for subject, subject_data in data.groupby('subject'):
            pool_lists = self.split_lists(subject_data, 'study', self.item_query)
            recall_lists = self.split_lists(subject_data, 'recall')
            results = self.analyze_subject(subject, pool_lists, recall_lists)
            subj_results.append(results)
        stat = pd.concat(subj_results, axis=0).reset_index()
        return stat


class TransitionOutputs(TransitionMeasure):
    """Measure recall probability by input and output position."""

    def __init__(self, list_length, item_query=None, test_key=None, test=None):
        super().__init__(
            'input', 'input', item_query=item_query, test_key=test_key, test=test
        )
        self.list_length = list_length

    def analyze_subject(self, subject, pool, recall):
        actual, possible = stats.count_outputs(
            self.list_length,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        input_pos = np.tile(np.arange(1, actual.shape[1] + 1), actual.shape[0])
        output_pos = np.repeat(np.arange(1, actual.shape[0] + 1), actual.shape[1])
        with np.errstate(divide='ignore', invalid='ignore'):
            prob = actual.flatten() / possible.flatten()
        pnr = pd.DataFrame(
            {
                'subject': subject,
                'input': input_pos,
                'output': output_pos,
                'prob': prob,
                'actual': actual.flatten(),
                'possible': possible.flatten(),
            }
        )
        pnr = pnr.set_index(['subject', 'output', 'input'])
        return pnr


class TransitionLag(TransitionMeasure):
    """Measure conditional response probability by lag."""

    def __init__(
        self,
        list_length,
        lag_key='input',
        count_unique=False,
        item_query=None,
        test_key=None,
        test=None,
        compound=False,
    ):
        super().__init__(
            'input', lag_key, item_query=item_query, test_key=test_key, test=test
        )
        self.list_length = list_length
        self.count_unique = count_unique
        self.compound = compound

    def analyze_subject(self, subject, pool, recall):

        if self.compound:
            counter = stats.count_lags_compound
        else:
            counter = stats.count_lags

        actual, possible = counter(
            self.list_length,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
            self.count_unique,
        )
        crp = pd.DataFrame(
            {
                'subject': subject,
                'prob': actual / possible,
                'actual': actual,
                'possible': possible,
            },
            index=actual.index,
        )
        if self.compound:
            crp = crp.set_index('subject', append=True)
            crp = crp.reorder_levels(['subject', 'previous', 'current'])
        else:
            crp = crp.set_index('subject', append=True)
            crp = crp.reorder_levels(['subject', 'lag'])
        return crp


class TransitionLagRank(TransitionMeasure):
    """Measure lag rank of transitions."""

    def __init__(self, lag_key='input', item_query=None, test_key=None, test=None):
        super().__init__(
            'input', lag_key, item_query=item_query, test_key=test_key, test=test
        )

    def analyze_subject(self, subject, pool, recall):
        ranks = stats.rank_lags(
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        stat = pd.DataFrame(
            {'subject': subject, 'rank': np.nanmean(ranks)}, index=[subject]
        )
        stat = stat.set_index('subject')
        return stat


class TransitionDistance(TransitionMeasure):
    """Measure conditional response probability by distance."""

    def __init__(
        self,
        index_key,
        distances,
        edges,
        count_unique=False,
        centers=None,
        item_query=None,
        test_key=None,
        test=None,
    ):
        super().__init__(
            'input', index_key, item_query=item_query, test_key=test_key, test=test
        )
        self.distances = distances
        self.edges = edges
        if centers is None:
            # if no explicit centers, use halfway between edges
            centers = edges[:-1] + (np.diff(edges) / 2)
        self.centers = centers
        self.count_unique = count_unique

    def analyze_subject(self, subject, pool, recall):

        actual, possible = stats.count_distance(
            self.distances,
            self.edges,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
            count_unique=self.count_unique,
        )
        crp = pd.DataFrame(
            {
                'subject': subject,
                'center': self.centers,
                'bin': actual.index,
                'prob': actual / possible,
                'actual': actual,
                'possible': possible,
            }
        )
        crp = crp.set_index(['subject', 'center'])
        return crp


class TransitionDistanceRank(TransitionMeasure):
    """Measure transition rank by distance."""

    def __init__(self, index_key, distances, item_query=None, test_key=None, test=None):
        super().__init__(
            index_key, index_key, item_query=item_query, test_key=test_key, test=test
        )
        self.distances = distances

    def analyze_subject(self, subject, pool, recall):
        ranks = stats.rank_distance(
            self.distances,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        stat = pd.DataFrame(
            {'subject': subject, 'rank': np.nanmean(ranks)}, index=[subject]
        )
        stat = stat.set_index('subject')
        return stat


class TransitionDistanceRankShifted(TransitionMeasure):
    """Measure shifted transition rank by distance."""

    def __init__(
        self, index_key, distances, max_shift, item_query=None, test_key=None, test=None
    ):
        super().__init__(
            index_key, index_key, item_query=item_query, test_key=test_key, test=test
        )
        self.distances = distances
        self.max_shift = int(max_shift)

    def analyze_subject(self, subject, pool, recall):
        ranks = stats.rank_distance_shifted(
            self.distances,
            self.max_shift,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        shifts = np.arange(-self.max_shift, 0)
        index = pd.MultiIndex.from_arrays(
            [[subject] * self.max_shift, shifts], names=['subject', 'shift']
        )
        stat = pd.DataFrame({'rank': np.nanmean(ranks, 0)}, index=index)
        return stat


class TransitionDistanceRankWindow(TransitionMeasure):
    """Measure transition distance rank within a window."""

    def __init__(
        self,
        index_key,
        distances,
        list_length,
        window_lags,
        item_query=None,
        test_key=None,
        test=None,
    ):
        super().__init__(
            'input', index_key, item_query=item_query, test_key=test_key, test=test
        )
        self.distances = distances
        self.list_length = list_length
        self.window_lags = window_lags

    def analyze_subject(self, subject, pool, recall):
        ranks = stats.rank_distance_window(
            self.distances,
            self.list_length,
            self.window_lags,
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        index = pd.MultiIndex.from_arrays(
            [[subject] * len(self.window_lags), self.window_lags],
            names=['subject', 'lag'],
        )
        stat = pd.DataFrame({'rank': np.nanmean(ranks, 0)}, index=index)
        return stat


class TransitionCategory(TransitionMeasure):
    """Measure conditional response probability by category transition."""

    def __init__(self, category_key, item_query=None, test_key=None, test=None):
        super().__init__(
            'input', category_key, item_query=item_query, test_key=test_key, test=test
        )

    def analyze_subject(self, subject, pool, recall):
        actual, possible = stats.count_category(
            pool['items'],
            recall['items'],
            pool['label'],
            recall['label'],
            pool['test'],
            recall['test'],
            self.test,
        )
        crp = pd.DataFrame(
            {
                'subject': subject,
                'prob': actual / possible,
                'actual': actual,
                'possible': possible,
            },
            index=[subject],
        )
        crp = crp.set_index('subject')
        return crp



"""Test category clusting statistics."""

import numpy as np
import pytest

from psifr import stats


@pytest.fixture()
def data():
    """Create list data with position and category."""
    test_list = {
        'list_length': 8,
        'pool_position': [1, 2, 3, 4, 5, 6, 7, 8],
        'pool_category': [1, 1, 1, 1, 2, 2, 2, 2],
        'output_position': [1, 3, 4, 8, 5, 4, 2, 7, 6],
        'output_category': [1, 1, 1, 2, 2, 1, 1, 2, 2],
    }
    return test_list


def test_category_count(data):
    """Test within category actual and possible counts."""
    # six transitions, but only five have a valid within-category
    # transition that is possible
    actual, possible = stats.count_category(
        [data['pool_position']],
        [data['output_position']],
        [data['pool_category']],
        [data['output_category']],
    )
    assert actual == 4
    assert possible == 5


@pytest.fixture()
def cases1():
    # test cases taken from Roenker et al. 1971, Psychological Bulletin
    recall_str = [
        'aaaaabacbbbbcbcccc',
        'aaaaabbbbccccabaaaacccccabcbbbbbccccccaaabbbbccaabbaba',
        'aacccccacbbcccccbb',
        'aaaaacbcccccccabbbcccccbcccccbbbbbcccaabcbccccccaaaacc',
        'aaaefbbbefcccedfdd',
        'aaaaadedbbbcbaaaaeeedbecccfffecccccddddedebffffffbbdeb',
        'abeeeecedffffffedf',
        'abbddddedfcccdeeeeedefdffffffcfffeefffffdceeeeeffffffd',
    ]
    data = {'recall': [np.array(list(s)) for s in recall_str]}
    return data


@pytest.fixture()
def cases2():
    # test cases taken from Stricker et al. 2002, J Int Neuropsychol Soc.
    # the CVLT isn't very well described in the sources I have, but
    # this study list matches the description (4 each of 4 categories,
    # same-category items not adjacent)
    study = [['a', 'b', 'c', 'd'] * 4] * 10
    recall_str = [
        'aabb',
        'aaaa',
        'aaaabbbb',
        'aabbccdd',
        'aaaabbbbcccc',
        'aaabbbcccddd',
        'aabbcdcd',
        'aaaacdcd',
        'aabbabab',
        'aaabbbcccdddabcd',
    ]
    recall = [np.array(list(s)) for s in recall_str]
    data = {'study': study, 'recall': recall}
    return data


def test_arc1(cases1):
    # checked the below expected ARC scores against the reported number of
    # clusters (R), expected clusters (E[R]), and maximum clusters (maxR).
    # The expected ARC scores also match the reported ones within rounding
    # error.
    expected = np.array([0.5, 0.5, 0.49295775, 0.5, 0.5, 0.5, 0.49295775, 0.5])
    observed = stats.arc(cases1['recall'])
    np.testing.assert_allclose(observed, expected)


def test_arc2(cases2):
    # original values below. One case was different from reported
    # expected = np.array([1.00, np.nan, 1.00, 1.00, 1.00, 1.00, 0.33, 0.33, 0.11, 0.55])
    expected = np.array([1.00, np.nan, 1.00, 1.00, 1.00, 1.00, 0.33, 0.33, -0.33, 0.55])
    observed = stats.arc(cases2['recall'])
    np.testing.assert_allclose(observed, expected, rtol=0.011)


def test_lbc(cases2):
    # original values below. Found some small deviations from the paper
    # for unclear reasons; probably rounding errors or typos
    # expected = np.array([1.40, 2.40, 4.62, 2.62, 6.80, 5.80, 0.62, 1.62, 0.62, 4.99])

    # corrected expected values for the Stricker test cases
    expected = np.array([1.40, 2.40, 4.60, 2.60, 6.80, 5.80, 0.60, 1.60, 0.60, 5.00])

    observed = stats.lbc(cases2['study'], cases2['recall'])
    np.testing.assert_allclose(observed, expected)
